#
# This code is part of Ansible, but is an independent component.
#
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# Copyright: (c) 2017, Red Hat Inc.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import absolute_import, division, print_function


__metaclass__ = type


import json
import re

from copy import deepcopy

from ansible.module_utils._text import to_text
from ansible.module_utils.common._collections_compat import Mapping
from ansible.module_utils.connection import Connection, ConnectionError
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.config import (
    CustomNetworkConfig,
    NetworkConfig,
    dumps,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    ComplexList,
    to_list,
)


try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False


_DEVICE_CONNECTION = None


def get_connection(module):
    global _DEVICE_CONNECTION
    if not _DEVICE_CONNECTION:
        connection_proxy = Connection(module._socket_path)
        cap = json.loads(connection_proxy.get_capabilities())
        if cap["network_api"] == "cliconf":
            conn = Cli(module)
        elif cap["network_api"] == "nxapi":
            conn = HttpApi(module)
        _DEVICE_CONNECTION = conn
    return _DEVICE_CONNECTION


class Cli:
    def __init__(self, module):
        self._module = module
        self._device_configs = {}
        self._connection = None

    def _get_connection(self):
        if self._connection:
            return self._connection
        self._connection = Connection(self._module._socket_path)

        return self._connection

    def get_config(self, flags=None):
        """Retrieves the current config from the device or cache"""
        flags = [] if flags is None else flags

        cmd = "show running-config "
        cmd += " ".join(flags)
        cmd = cmd.strip()

        try:
            return self._device_configs[cmd]
        except KeyError:
            connection = self._get_connection()
            try:
                out = connection.get_config(flags=flags)
            except ConnectionError as exc:
                self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

            cfg = to_text(out, errors="surrogate_then_replace").strip() + "\n"
            self._device_configs[cmd] = cfg
            return cfg

    def run_commands(self, commands, check_rc=True):
        """Run list of commands on remote device and return results"""
        connection = self._get_connection()

        try:
            out = connection.run_commands(commands, check_rc)
            if check_rc == "retry_json":
                capabilities = self.get_capabilities()
                network_api = capabilities.get("network_api")

                if network_api == "cliconf" and out:
                    for index, resp in enumerate(out):
                        if (
                            "Invalid command at" in resp or "Ambiguous command at" in resp
                        ) and "json" in resp:
                            if commands[index]["output"] == "json":
                                commands[index]["output"] = "text"
                                out = connection.run_commands(commands, check_rc)
            return out
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc))

    def load_config(self, config, return_error=False, opts=None, replace=None):
        """Sends configuration commands to the remote device"""
        if opts is None:
            opts = {}

        connection = self._get_connection()
        responses = []
        try:
            resp = connection.edit_config(config, replace=replace)
            if isinstance(resp, Mapping):
                resp = resp["response"]
        except ConnectionError as e:
            code = getattr(e, "code", 1)
            message = getattr(e, "err", e)
            err = to_text(message, errors="surrogate_then_replace")
            if opts.get("ignore_timeout") and code:
                responses.append(err)
                return responses
            elif code and "no graceful-restart" in err:
                if "ISSU/HA will be affected if Graceful Restart is disabled" in err:
                    msg = [""]
                    responses.extend(msg)
                    return responses
                else:
                    self._module.fail_json(msg=err)
            elif code:
                self._module.fail_json(msg=err)

        responses.extend(resp)
        return responses

    def get_diff(
        self,
        candidate=None,
        running=None,
        diff_match="line",
        diff_ignore_lines=None,
        path=None,
        diff_replace="line",
    ):
        conn = self._get_connection()
        try:
            response = conn.get_diff(
                candidate=candidate,
                running=running,
                diff_match=diff_match,
                diff_ignore_lines=diff_ignore_lines,
                path=path,
                diff_replace=diff_replace,
            )
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))
        return response

    def get_capabilities(self):
        """Returns platform info of the remove device"""
        if hasattr(self._module, "_capabilities"):
            return self._module._capabilities

        connection = self._get_connection()
        try:
            capabilities = connection.get_capabilities()
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))
        self._module._capabilities = json.loads(capabilities)
        return self._module._capabilities

    def read_module_context(self, module_key):
        connection = self._get_connection()
        try:
            module_context = connection.read_module_context(module_key)
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

        return module_context

    def save_module_context(self, module_key, module_context):
        connection = self._get_connection()
        try:
            connection.save_module_context(module_key, module_context)
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

        return None


class HttpApi:
    def __init__(self, module):
        self._module = module
        self._device_configs = {}
        self._module_context = {}
        self._connection_obj = None

    @property
    def _connection(self):
        if not self._connection_obj:
            self._connection_obj = Connection(self._module._socket_path)

        return self._connection_obj

    def run_commands(self, commands, check_rc=True):
        """Runs list of commands on remote device and returns results"""
        try:
            out = self._connection.send_request(commands)
        except ConnectionError as exc:
            if check_rc is True:
                raise
            out = to_text(exc)

        out = to_list(out)
        if not out[0]:
            return out

        for index, response in enumerate(out):
            if response[0] == "{":
                out[index] = json.loads(response)

        return out

    def get_config(self, flags=None):
        """Retrieves the current config from the device or cache"""
        flags = [] if flags is None else flags

        cmd = "show running-config "
        cmd += " ".join(flags)
        cmd = cmd.strip()

        try:
            return self._device_configs[cmd]
        except KeyError:
            try:
                out = self._connection.send_request(cmd)
            except ConnectionError as exc:
                self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

            cfg = to_text(out).strip()
            self._device_configs[cmd] = cfg
            return cfg

    def get_diff(
        self,
        candidate=None,
        running=None,
        diff_match="line",
        diff_ignore_lines=None,
        path=None,
        diff_replace="line",
    ):
        diff = {}

        # prepare candidate configuration
        candidate_obj = NetworkConfig(indent=2)
        candidate_obj.load(candidate)

        if running and diff_match != "none" and diff_replace != "config":
            # running configuration
            running_obj = NetworkConfig(indent=2, contents=running, ignore_lines=diff_ignore_lines)
            configdiffobjs = candidate_obj.difference(
                running_obj,
                path=path,
                match=diff_match,
                replace=diff_replace,
            )

        else:
            configdiffobjs = candidate_obj.items

        diff["config_diff"] = dumps(configdiffobjs, "commands") if configdiffobjs else ""
        return diff

    def load_config(self, commands, return_error=False, opts=None, replace=None):
        """Sends the ordered set of commands to the device"""
        if opts is None:
            opts = {}

        responses = []
        try:
            resp = self.edit_config(commands, replace=replace)
        except ConnectionError as exc:
            code = getattr(exc, "code", 1)
            message = getattr(exc, "err", exc)
            err = to_text(message, errors="surrogate_then_replace")
            if opts.get("ignore_timeout") and code:
                responses.append(code)
                return responses
            elif opts.get("catch_clierror") and "400" in code:
                return [code, err]
            elif code and "no graceful-restart" in err:
                if "ISSU/HA will be affected if Graceful Restart is disabled" in err:
                    msg = [""]
                    responses.extend(msg)
                    return responses
                else:
                    self._module.fail_json(msg=err)
            elif code:
                self._module.fail_json(msg=err)

        responses.extend(resp)
        return responses

    def edit_config(self, candidate=None, commit=True, replace=None, comment=None):
        resp = list()

        self.check_edit_config_capability(candidate, commit, replace, comment)

        if replace:
            candidate = "config replace {0}".format(replace)

        responses = self._connection.send_request(candidate, output="config")
        for response in to_list(responses):
            if response != "{}":
                resp.append(response)
        if not resp:
            resp = [""]

        return resp

    def get_capabilities(self):
        """Returns platform info of the remove device"""
        try:
            capabilities = self._connection.get_capabilities()
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

        return json.loads(capabilities)

    def check_edit_config_capability(self, candidate=None, commit=True, replace=None, comment=None):
        operations = self._connection.get_device_operations()

        if not candidate and not replace:
            raise ValueError("must provide a candidate or replace to load configuration")

        if commit not in (True, False):
            raise ValueError("'commit' must be a bool, got %s" % commit)

        if replace and not operations.get("supports_replace"):
            raise ValueError("configuration replace is not supported")

        if comment and not operations.get("supports_commit_comment", False):
            raise ValueError("commit comment is not supported")

    def read_module_context(self, module_key):
        try:
            module_context = self._connection.read_module_context(module_key)
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

        return module_context

    def save_module_context(self, module_key, module_context):
        try:
            self._connection.save_module_context(module_key, module_context)
        except ConnectionError as exc:
            self._module.fail_json(msg=to_text(exc, errors="surrogate_then_replace"))

        return None


class NxosCmdRef:
    """NXOS Command Reference utilities.
    The NxosCmdRef class takes a yaml-formatted string of nxos module commands
    and converts it into dict-formatted database of getters/setters/defaults
    and associated common and platform-specific values. The utility methods
    add additional data such as existing states, playbook states, and proposed cli.
    The utilities also abstract away platform differences such as different
    defaults and different command syntax.

    Callers must provide a yaml formatted string that defines each command and
    its properties; e.g. BFD global:
    ---
    _template: # _template holds common settings for all commands
      # Enable feature bfd if disabled
      feature: bfd
      # Common getter syntax for BFD commands
      get_command: show run bfd all | incl '^(no )*bfd'

    interval:
      kind: dict
      getval: bfd interval (?P<tx>\\d+) min_rx (?P<min_rx>\\d+) multiplier (?P<multiplier>\\d+)
      setval: bfd interval {tx} min_rx {min_rx} multiplier {multiplier}
      default:
        tx: 50
        min_rx: 50
        multiplier: 3
      N3K:
        # Platform overrides
        default:
          tx: 250
          min_rx: 250
          multiplier: 3
    """

    def __init__(self, module, cmd_ref_str, ref_only=False):
        """Initialize cmd_ref from yaml data."""

        self._module = module
        self._check_imports()
        self._yaml_load(cmd_ref_str)
        self.cache_existing = None
        self.present_states = ["present", "merged", "replaced"]
        self.absent_states = ["absent", "deleted"]
        ref = self._ref

        # Create a list of supported commands based on ref keys
        ref["commands"] = sorted([k for k in ref if not k.startswith("_")])
        ref["_proposed"] = []
        ref["_context"] = []
        ref["_resource_key"] = None

        if not ref_only:
            ref["_state"] = module.params.get("state", "present")
            self.feature_enable()
            self.get_platform_defaults()
            self.normalize_defaults()

    def __getitem__(self, key=None):
        if key is None:
            return self._ref
        return self._ref[key]

    def _check_imports(self):
        module = self._module
        msg = nxosCmdRef_import_check()
        if msg:
            module.fail_json(msg=msg)

    def _yaml_load(self, cmd_ref_str):
        self._ref = yaml.load(cmd_ref_str, Loader=yaml.FullLoader)

    def feature_enable(self):
        """Add 'feature <foo>' to _proposed if ref includes a 'feature' key."""
        ref = self._ref
        feature = ref["_template"].get("feature")
        if feature:
            show_cmd = "show run | incl 'feature {0}'".format(feature)
            output = self.execute_show_command(show_cmd, "text")
            if not output or "CLI command error" in output:
                msg = "** 'feature {0}' is not enabled. Module will auto-enable feature {0} ** ".format(
                    feature,
                )
                self._module.warn(msg)
                ref["_proposed"].append("feature {0}".format(feature))
                ref["_cli_is_feature_disabled"] = ref["_proposed"]

    def get_platform_shortname(self):
        """Query device for platform type, normalize to a shortname/nickname.
        Returns platform shortname (e.g. 'N3K-3058P' returns 'N3K') or None.
        """
        # TBD: add this method logic to get_capabilities() after those methods
        #      are made consistent across transports
        platform_info = self.execute_show_command("show inventory", "json")
        if not platform_info or not isinstance(platform_info, dict):
            return None
        inventory_table = platform_info["TABLE_inv"]["ROW_inv"]
        for info in inventory_table:
            if "Chassis" in info["name"]:
                network_os_platform = info["productid"]
                break
        else:
            return None

        # Supported Platforms: N3K,N5K,N6K,N7K,N9K,N3K-F,N9K-F
        m = re.match("(?P<short>N[35679][K57])-(?P<N35>C35)*", network_os_platform)
        if not m:
            return None
        shortname = m.group("short")

        # Normalize
        if m.groupdict().get("N35"):
            shortname = "N35"
        elif re.match("N77", shortname):
            shortname = "N7K"
        elif re.match(r"N3K|N9K", shortname):
            for info in inventory_table:
                if "-R" in info["productid"]:
                    # Fretta Platform
                    shortname += "-F"
                    break
        return shortname

    def get_platform_defaults(self):
        """Update ref with platform specific defaults"""
        plat = self.get_platform_shortname()
        if not plat:
            return

        ref = self._ref
        ref["_platform_shortname"] = plat
        # Remove excluded commands (no platform support for command)
        for k in ref["commands"]:
            if plat in ref[k].get("_exclude", ""):
                ref["commands"].remove(k)

        # Update platform-specific settings for each item in ref
        plat_spec_cmds = [k for k in ref["commands"] if plat in ref[k]]
        for k in plat_spec_cmds:
            for plat_key in ref[k][plat]:
                ref[k][plat_key] = ref[k][plat][plat_key]

    def normalize_defaults(self):
        """Update ref defaults with normalized data"""
        ref = self._ref
        for k in ref["commands"]:
            if "default" in ref[k] and ref[k]["default"]:
                kind = ref[k]["kind"]
                if "int" == kind:
                    ref[k]["default"] = int(ref[k]["default"])
                elif "list" == kind:
                    ref[k]["default"] = [str(i) for i in ref[k]["default"]]
                elif "dict" == kind:
                    for key, v in ref[k]["default"].items():
                        if v:
                            v = str(v)
                        ref[k]["default"][key] = v

    def execute_show_command(self, command, format):
        """Generic show command helper.
        Warning: 'CLI command error' exceptions are caught, must be handled by caller.
        Return device output as a newline-separated string or None.
        """
        cmds = [{"command": command, "output": format}]
        output = None
        try:
            output = run_commands(self._module, cmds)
            if output:
                output = output[0]
        except ConnectionError as exc:
            if "CLI command error" in repr(exc):
                # CLI may be feature disabled
                output = repr(exc)
            else:
                raise
        return output

    def pattern_match_existing(self, output, k):
        """Pattern matching helper for `get_existing`.
        `k` is the command name string. Use the pattern from cmd_ref to
        find a matching string in the output.
        Return regex match object or None.
        """
        ref = self._ref
        pattern = re.compile(ref[k]["getval"])
        multiple = "multiple" in ref[k].keys()
        match_lines = [re.search(pattern, line) for line in output]
        if "dict" == ref[k]["kind"]:
            match = [m for m in match_lines if m]
            if not match:
                return None
            if len(match) > 1 and not multiple:
                raise ValueError("get_existing: multiple matches found for property {0}".format(k))
        else:
            match = [m.groups() for m in match_lines if m]
            if not match:
                return None
            if len(match) > 1 and not multiple:
                raise ValueError("get_existing: multiple matches found for property {0}".format(k))
            for item in match:
                index = match.index(item)
                match[index] = list(item)  # tuple to list

                # Handle config strings that nvgen with the 'no' prefix.
                # Example match behavior:
                # When pattern is: '(no )*foo *(\S+)*$' AND
                #  When output is: 'no foo'  -> match: ['no ', None]
                #  When output is: 'foo 50'  -> match: [None, '50']
                if None is match[index][0]:
                    match[index].pop(0)
                elif "no" in match[index][0]:
                    match[index].pop(0)
                    if not match:
                        return None

        return match

    def set_context(self, context=None):
        """Update ref with command context."""
        if context is None:
            context = []
        ref = self._ref
        # Process any additional context that this propoerty might require.
        # 1) Global context from NxosCmdRef _template.
        # 2) Context passed in using context arg.
        ref["_context"] = ref["_template"].get("context", [])
        for cmd in context:
            ref["_context"].append(cmd)
        # Last key in context is the resource key
        ref["_resource_key"] = context[-1] if context else ref["_resource_key"]

    def get_existing(self, cache_output=None):
        """Update ref with existing command states from the device.
        Store these states in each command's 'existing' key.
        """
        ref = self._ref
        if ref.get("_cli_is_feature_disabled"):
            # Add context to proposed if state is present
            if ref["_state"] in self.present_states:
                [ref["_proposed"].append(ctx) for ctx in ref["_context"]]
            return

        show_cmd = ref["_template"]["get_command"]
        if cache_output:
            output = cache_output
        else:
            output = self.execute_show_command(show_cmd, "text") or []
            self.cache_existing = output

        # Add additional command context if needed.
        if ref["_context"]:
            output = CustomNetworkConfig(indent=2, contents=output)
            output = output.get_section(ref["_context"])

        if not output:
            # Add context to proposed if state is present
            if ref["_state"] in self.present_states:
                [ref["_proposed"].append(ctx) for ctx in ref["_context"]]
            return

        # We need to remove the last item in context for state absent case.
        if ref["_state"] in self.absent_states and ref["_context"]:
            if ref["_resource_key"] and ref["_resource_key"] == ref["_context"][-1]:
                if ref["_context"][-1] in output:
                    ref["_context"][-1] = "no " + ref["_context"][-1]
                else:
                    del ref["_context"][-1]
                return

        # Walk each cmd in ref, use cmd pattern to discover existing cmds
        output = output.split("\n")
        for k in ref["commands"]:
            match = self.pattern_match_existing(output, k)
            if not match:
                continue
            ref[k]["existing"] = {}
            for item in match:
                index = match.index(item)
                kind = ref[k]["kind"]
                if "int" == kind:
                    ref[k]["existing"][index] = int(item[0])
                elif "list" == kind:
                    ref[k]["existing"][index] = [str(i) for i in item[0]]
                elif "dict" == kind:
                    # The getval pattern should contain regex named group keys that
                    # match up with the setval named placeholder keys; e.g.
                    #   getval: my-cmd (?P<foo>\d+) bar (?P<baz>\d+)
                    #   setval: my-cmd {foo} bar {baz}
                    ref[k]["existing"][index] = {}
                    for key in item.groupdict().keys():
                        ref[k]["existing"][index][key] = str(item.group(key))
                elif "str" == kind:
                    ref[k]["existing"][index] = item[0]
                else:
                    raise ValueError(
                        "get_existing: unknown 'kind' value specified for key '{0}'".format(k),
                    )

    def get_playvals(self):
        """Update ref with values from the playbook.
        Store these values in each command's 'playval' key.
        """
        ref = self._ref
        module = self._module
        params = {}
        if module.params.get("config"):
            # Resource module builder packs playvals under 'config' key
            param_data = module.params.get("config")
            params["global"] = param_data
            for key in param_data.keys():
                if isinstance(param_data[key], list):
                    params[key] = param_data[key]
        else:
            params["global"] = module.params
        for k in ref.keys():
            for level in params.keys():
                if isinstance(params[level], dict):
                    params[level] = [params[level]]
                for item in params[level]:
                    if k in item and item[k] is not None:
                        if not ref[k].get("playval"):
                            ref[k]["playval"] = {}
                        playval = item[k]
                        index = params[level].index(item)
                        # Normalize each value
                        if "int" == ref[k]["kind"]:
                            playval = int(playval)
                        elif "list" == ref[k]["kind"]:
                            playval = [str(i) for i in playval]
                        elif "dict" == ref[k]["kind"]:
                            for key, v in playval.items():
                                playval[key] = str(v)
                        ref[k]["playval"][index] = playval

    def build_cmd_set(self, playval, existing, k):
        """Helper function to create list of commands to configure device
        Return a list of commands
        """
        ref = self._ref
        proposed = ref["_proposed"]
        cmd = None
        kind = ref[k]["kind"]
        if "int" == kind:
            cmd = ref[k]["setval"].format(playval)
        elif "list" == kind:
            cmd = ref[k]["setval"].format(*(playval))
        elif "dict" == kind:
            # The setval pattern should contain placeholder keys that
            # match up with the getval regex named group keys; e.g.
            #   getval: my-cmd (?P<foo>\d+) bar (?P<baz>\d+)
            #   setval: my-cmd {foo} bar {baz}
            if ref[k]["setval"].startswith("path"):
                tmplt = "path {name}"
                if "depth" in playval:
                    tmplt += " depth {depth}"
                if "query_condition" in playval:
                    tmplt += " query-condition {query_condition}"
                if "filter_condition" in playval:
                    tmplt += " filter-condition {filter_condition}"
                cmd = tmplt.format(**playval)
            else:
                cmd = ref[k]["setval"].format(**playval)
        elif "str" == kind:
            if "deleted" in str(playval):
                if existing:
                    cmd = "no " + ref[k]["setval"].format(existing)
            else:
                cmd = ref[k]["setval"].format(playval)
        else:
            raise ValueError("get_proposed: unknown 'kind' value specified for key '{0}'".format(k))
        if cmd:
            if ref["_state"] in self.absent_states and not re.search(r"^no", cmd):
                cmd = "no " + cmd
            # Commands may require parent commands for proper context.
            # Global _template context is replaced by parameter context
            [proposed.append(ctx) for ctx in ref["_context"]]
            [proposed.append(ctx) for ctx in ref[k].get("context", [])]
            proposed.append(cmd)

    def get_proposed(self):
        """Compare playbook values against existing states and create a list
        of proposed commands.
        Return a list of raw cli command strings.
        """
        ref = self._ref
        # '_proposed' may be empty list or contain initializations; e.g. ['feature foo']
        proposed = ref["_proposed"]

        if ref["_context"] and ref["_context"][-1].startswith("no"):
            [proposed.append(ctx) for ctx in ref["_context"]]
            return proposed

        # Create a list of commands that have playbook values
        play_keys = [k for k in ref["commands"] if "playval" in ref[k]]

        def compare(playval, existing):
            if ref["_state"] in self.present_states:
                if existing is None:
                    return False
                elif str(playval) == str(existing):
                    return True
                elif isinstance(existing, dict) and playval in existing.values():
                    return True

            if ref["_state"] in self.absent_states:
                if isinstance(existing, dict) and all(x is None for x in existing.values()):
                    existing = None
                if existing is None or playval not in existing.values():
                    return True
            return False

        # Compare against current state
        for k in play_keys:
            playval = ref[k]["playval"]
            # Create playval copy to avoid RuntimeError
            #   dictionary changed size during iteration error
            playval_copy = deepcopy(playval)
            existing = ref[k].get("existing", ref[k]["default"])
            multiple = "multiple" in ref[k].keys()

            # Multiple Instances:
            if isinstance(existing, dict) and multiple:
                for ekey, evalue in existing.items():
                    if isinstance(evalue, dict):
                        # Remove values set to string 'None' from dvalue
                        evalue = dict((k, v) for k, v in evalue.items() if v != "None")
                    for pkey, pvalue in playval.items():
                        if compare(pvalue, evalue):
                            if playval_copy.get(pkey):
                                del playval_copy[pkey]
                if not playval_copy:
                    continue
            # Single Instance:
            else:
                for pkey, pval in playval.items():
                    if compare(pval, existing):
                        if playval_copy.get(pkey):
                            del playval_copy[pkey]
                if not playval_copy:
                    continue

            playval = playval_copy
            # Multiple Instances:
            if isinstance(existing, dict):
                for dkey, dvalue in existing.items():
                    for pval in playval.values():
                        self.build_cmd_set(pval, dvalue, k)
            # Single Instance:
            else:
                for pval in playval.values():
                    self.build_cmd_set(pval, existing, k)

        # Remove any duplicate commands before returning.
        # pylint: disable=unnecessary-lambda
        cmds = sorted(set(proposed), key=lambda x: proposed.index(x))
        return cmds


def nxosCmdRef_import_check():
    """Return import error messages or empty string"""
    msg = ""
    if not HAS_YAML:
        msg += "Mandatory python library 'PyYAML' is not present, try 'pip install PyYAML'\n"
    return msg


def is_json(cmd):
    return to_text(cmd).endswith("| json")


def is_text(cmd):
    return not is_json(cmd)


def to_command(module, commands):
    transform = ComplexList(
        dict(
            command=dict(key=True),
            output=dict(type="str", default="text"),
            prompt=dict(type="list"),
            answer=dict(type="list"),
            newline=dict(type="bool", default=True),
            sendonly=dict(type="bool", default=False),
            check_all=dict(type="bool", default=False),
        ),
        module,
    )

    commands = transform(to_list(commands))

    for item in commands:
        if is_json(item["command"]):
            item["output"] = "json"

    return commands


def get_config(module, flags=None):
    flags = [] if flags is None else flags

    conn = get_connection(module)
    return conn.get_config(flags=flags)


def run_commands(module, commands, check_rc=True):
    conn = get_connection(module)
    return conn.run_commands(to_command(module, commands), check_rc)


def load_config(module, config, return_error=False, opts=None, replace=None):
    conn = get_connection(module)
    return conn.load_config(config, return_error, opts, replace=replace)


def get_capabilities(module):
    conn = get_connection(module)
    return conn.get_capabilities()


def get_diff(
    self,
    candidate=None,
    running=None,
    diff_match="line",
    diff_ignore_lines=None,
    path=None,
    diff_replace="line",
):
    conn = self.get_connection()
    return conn.get_diff(
        candidate=candidate,
        running=running,
        diff_match=diff_match,
        diff_ignore_lines=diff_ignore_lines,
        path=path,
        diff_replace=diff_replace,
    )


def normalize_interface(name):
    """Return the normalized interface name"""
    if not name:
        return

    def _get_number(name):
        digits = ""
        for char in name:
            if char.isdigit() or char in "/.":
                digits += char
        return digits

    if name.lower().startswith("et"):
        if_type = "Ethernet"
    elif name.lower().startswith("vl"):
        if_type = "Vlan"
    elif name.lower().startswith("lo"):
        if_type = "loopback"
    elif name.lower().startswith("po"):
        if_type = "port-channel"
    elif name.lower().startswith("nv"):
        if_type = "nve"
    else:
        if_type = None

    number_list = name.split(" ")
    if len(number_list) == 2:
        number = number_list[-1].strip()
    else:
        number = _get_number(name)

    if if_type:
        proper_interface = if_type + number
    else:
        proper_interface = name

    return proper_interface


def get_interface_type(interface):
    """Gets the type of interface"""
    if interface.upper().startswith("ET"):
        return "ethernet"
    elif interface.upper().startswith("VL"):
        return "svi"
    elif interface.upper().startswith("LO"):
        return "loopback"
    elif interface.upper().startswith("MG"):
        return "management"
    elif interface.upper().startswith("MA"):
        return "management"
    elif interface.upper().startswith("PO"):
        return "portchannel"
    elif interface.upper().startswith("NV"):
        return "nve"
    else:
        return "unknown"


def default_intf_enabled(name="", sysdefs=None, mode=None):
    """Get device/version/interface-specific default 'enabled' state.
    L3:
     - Most L3 intfs default to 'shutdown'. Loopbacks default to 'no shutdown'.
     - Some legacy platforms default L3 intfs to 'no shutdown'.
    L2:
     - User-System-Default 'system default switchport shutdown' defines the
       enabled state for L2 intf's. USD defaults may be different on some platforms.
     - An intf may be explicitly defined as L2 with 'switchport' or it may be
       implicitly defined as L2 when USD 'system default switchport' is defined.
    """
    if not name:
        return None
    if sysdefs is None:
        sysdefs = {}
    default = False

    if re.search("port-channel|loopback", name):
        default = True
    elif re.search("Vlan", name):
        default = False
    else:
        if mode is None:
            # intf 'switchport' cli is not present so use the user-system-default
            mode = sysdefs.get("mode")

        if mode == "layer3":
            default = sysdefs.get("L3_enabled")
        elif mode == "layer2":
            default = sysdefs.get("L2_enabled")
    return default


def read_module_context(module):
    conn = get_connection(module)
    return conn.read_module_context(module._name)


def save_module_context(module, module_context):
    conn = get_connection(module)
    return conn.save_module_context(module._name, module_context)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: elasticache_parameter_group
version_added: 1.0.0
short_description: Manage cache parameter groups in Amazon ElastiCache.
description:
  - Manage cache security groups in Amazon ElastiCache.
  - Returns information about the specified cache cluster.
author:
  - "Sloane Hertel (@s-hertel)"

options:
  group_family:
    description:
      - The name of the cache parameter group family that the cache parameter group can be used with.
        Required when creating a cache parameter group.
    choices: ['memcached1.4', 'memcached1.5', 'redis2.6', 'redis2.8', 'redis3.2', 'redis4.0', 'redis5.0', 'redis6.x']
    type: str
  name:
    description:
     - A user-specified name for the cache parameter group.
    required: true
    type: str
  description:
    description:
      - A user-specified description for the cache parameter group.
    type: str
    default: ''
  state:
    description:
      - Idempotent actions that will create/modify, destroy, or reset a cache parameter group as needed.
    choices: ['present', 'absent', 'reset']
    required: true
    type: str
  values:
    description:
      - A user-specified dictionary of parameters to reset or modify for the cache parameter group.
    type: dict

extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.
- name: Create, modify and delete a parameter group
  hosts: localhost
  connection: local
  tasks:
    - name: 'Create a test parameter group'
      community.aws.elasticache_parameter_group:
        name: 'test-param-group'
        group_family: 'redis3.2'
        description: 'This is a cache parameter group'
        state: 'present'
    - name: 'Modify a test parameter group'
      community.aws.elasticache_parameter_group:
        name: 'test-param-group'
        values:
          activerehashing: true
          client-output-buffer-limit-normal-hard-limit: 4
        state: 'present'
    - name: 'Reset all modifiable parameters for the test parameter group'
      community.aws.elasticache_parameter_group:
        name: 'test-param-group'
        state: reset
    - name: 'Delete a test parameter group'
      community.aws.elasticache_parameter_group:
        name: 'test-param-group'
        state: 'absent'
"""

RETURN = r"""
elasticache:
  description: cache parameter group information and response metadata
  returned: always
  type: dict
  sample:
    cache_parameter_group:
      cache_parameter_group_family: redis3.2
      cache_parameter_group_name: test-please-delete
      description: "initial description"
    response_metadata:
      http_headers:
        content-length: "562"
        content-type: text/xml
        date: "Mon, 06 Feb 2017 22:14:08 GMT"
        x-amzn-requestid: 947291f9-ecb9-11e6-85bd-3baa4eca2cc1
      http_status_code: 200
      request_id: 947291f9-ecb9-11e6-85bd-3baa4eca2cc1
      retry_attempts: 0
changed:
  description: if the cache parameter group has changed
  returned: always
  type: bool
  sample:
    changed: true
"""

try:
    import botocore
except ImportError:
    pass  # Handled by AnsibleAWSModule

from ansible.module_utils._text import to_text
from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def create(module, conn, name, group_family, description):
    """Create ElastiCache parameter group."""
    try:
        response = conn.create_cache_parameter_group(
            CacheParameterGroupName=name, CacheParameterGroupFamily=group_family, Description=description
        )
        changed = True
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Unable to create cache parameter group.")
    return response, changed


def delete(module, conn, name):
    """Delete ElastiCache parameter group."""
    try:
        conn.delete_cache_parameter_group(CacheParameterGroupName=name)
        response = {}
        changed = True
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Unable to delete cache parameter group.")
    return response, changed


def make_current_modifiable_param_dict(module, conn, name):
    """Gets the current state of the cache parameter group and creates a dict with the format: {ParameterName: [Allowed_Values, DataType, ParameterValue]}"""
    current_info = get_info(conn, name)
    if current_info is False:
        module.fail_json(msg=f"Could not connect to the cache parameter group {name}.")

    parameters = current_info["Parameters"]
    modifiable_params = {}

    for param in parameters:
        if param["IsModifiable"]:
            modifiable_params[param["ParameterName"]] = [param.get("AllowedValues")]
            modifiable_params[param["ParameterName"]].append(param["DataType"])
            modifiable_params[param["ParameterName"]].append(param.get("ParameterValue"))
    return modifiable_params


def check_valid_modification(module, values, modifiable_params):
    """Check if the parameters and values in values are valid."""
    changed_with_update = False

    for parameter in values:
        new_value = values[parameter]

        # check valid modifiable parameters
        if parameter not in modifiable_params:
            module.fail_json(
                msg=f"{parameter} is not a modifiable parameter. Valid parameters to modify are: {modifiable_params.keys()}."
            )

        # check allowed datatype for modified parameters
        str_to_type = {"integer": int, "string": str}
        expected_type = str_to_type[modifiable_params[parameter][1]]
        if not isinstance(new_value, expected_type):
            if expected_type is str:
                if isinstance(new_value, bool):
                    values[parameter] = "yes" if new_value else "no"
                else:
                    values[parameter] = to_text(new_value)
            elif expected_type is int:
                if isinstance(new_value, bool):
                    values[parameter] = 1 if new_value else 0
                else:
                    module.fail_json(
                        msg=(
                            f"{new_value} (type {type(new_value)}) is not an allowed value for the parameter"
                            f" {parameter}. Expected a type {modifiable_params[parameter][1]}."
                        )
                    )
            else:
                module.fail_json(
                    msg=(
                        f"{new_value} (type {type(new_value)}) is not an allowed value for the parameter {parameter}."
                        f" Expected a type {modifiable_params[parameter][1]}."
                    )
                )

        # check allowed values for modifiable parameters
        choices = modifiable_params[parameter][0]
        if choices:
            if not (to_text(new_value) in choices or isinstance(new_value, int)):
                module.fail_json(
                    msg=f"{new_value} is not an allowed value for the parameter {parameter}. Valid parameters are: {choices}."
                )

        # check if a new value is different from current value
        if to_text(values[parameter]) != modifiable_params[parameter][2]:
            changed_with_update = True

    return changed_with_update, values


def check_changed_parameter_values(values, old_parameters, new_parameters):
    """Checking if the new values are different than the old values."""
    changed_with_update = False

    # if the user specified parameters to reset, only check those for change
    if values:
        for parameter in values:
            if old_parameters[parameter] != new_parameters[parameter]:
                changed_with_update = True
                break
    # otherwise check all to find a change
    else:
        for parameter in old_parameters:
            if old_parameters[parameter] != new_parameters[parameter]:
                changed_with_update = True
                break

    return changed_with_update


def modify(module, conn, name, values):
    """Modify ElastiCache parameter group to reflect the new information if it differs from the current."""
    # compares current group parameters with the parameters we've specified to to a value to see if this will change the group
    format_parameters = []
    for key in values:
        value = to_text(values[key])
        format_parameters.append({"ParameterName": key, "ParameterValue": value})
    try:
        response = conn.modify_cache_parameter_group(
            CacheParameterGroupName=name, ParameterNameValues=format_parameters
        )
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Unable to modify cache parameter group.")
    return response


def reset(module, conn, name, values):
    """Reset ElastiCache parameter group if the current information is different from the new information."""
    # used to compare with the reset parameters' dict to see if there have been changes
    old_parameters_dict = make_current_modifiable_param_dict(module, conn, name)

    format_parameters = []

    # determine whether to reset all or specific parameters
    if values:
        all_parameters = False
        format_parameters = []
        for key in values:
            value = to_text(values[key])
            format_parameters.append({"ParameterName": key, "ParameterValue": value})
    else:
        all_parameters = True

    try:
        response = conn.reset_cache_parameter_group(
            CacheParameterGroupName=name, ParameterNameValues=format_parameters, ResetAllParameters=all_parameters
        )
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Unable to reset cache parameter group.")

    # determine changed
    new_parameters_dict = make_current_modifiable_param_dict(module, conn, name)
    changed = check_changed_parameter_values(values, old_parameters_dict, new_parameters_dict)

    return response, changed


def get_info(conn, name):
    """Gets info about the ElastiCache parameter group. Returns false if it doesn't exist or we don't have access."""
    try:
        data = conn.describe_cache_parameters(CacheParameterGroupName=name)
        return data
    except botocore.exceptions.ClientError:
        return False


def main():
    argument_spec = dict(
        group_family=dict(
            type="str",
            choices=[
                "memcached1.4",
                "memcached1.5",
                "redis2.6",
                "redis2.8",
                "redis3.2",
                "redis4.0",
                "redis5.0",
                "redis6.x",
            ],
        ),
        name=dict(required=True, type="str"),
        description=dict(default="", type="str"),
        state=dict(required=True, choices=["present", "absent", "reset"]),
        values=dict(type="dict"),
    )
    module = AnsibleAWSModule(argument_spec=argument_spec)

    parameter_group_family = module.params.get("group_family")
    parameter_group_name = module.params.get("name")
    group_description = module.params.get("description")
    state = module.params.get("state")
    values = module.params.get("values")

    try:
        connection = module.client("elasticache")
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed to connect to AWS")

    exists = get_info(connection, parameter_group_name)

    # check that the needed requirements are available
    if state == "present" and not (exists or parameter_group_family):
        module.fail_json(msg="Creating a group requires a family group.")
    elif state == "reset" and not exists:
        module.fail_json(
            msg=f"No group {parameter_group_name} to reset. Please create the group before using the state 'reset'."
        )

    # Taking action
    changed = False
    if state == "present":
        if exists:
            # confirm that the group exists without any actions
            if not values:
                response = exists
                changed = False
            # modify existing group
            else:
                modifiable_params = make_current_modifiable_param_dict(module, connection, parameter_group_name)
                changed, values = check_valid_modification(module, values, modifiable_params)
                response = modify(module, connection, parameter_group_name, values)
        # create group
        else:
            response, changed = create(
                module, connection, parameter_group_name, parameter_group_family, group_description
            )
            if values:
                modifiable_params = make_current_modifiable_param_dict(module, connection, parameter_group_name)
                changed, values = check_valid_modification(module, values, modifiable_params)
                response = modify(module, connection, parameter_group_name, values)
    elif state == "absent":
        if exists:
            # delete group
            response, changed = delete(module, connection, parameter_group_name)
        else:
            response = {}
            changed = False
    elif state == "reset":
        response, changed = reset(module, connection, parameter_group_name, values)

    facts_result = dict(changed=changed, elasticache=camel_dict_to_snake_dict(response))

    module.exit_json(**facts_result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2024, Alexei Znamensky <russoz@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: django_check
author:
  - Alexei Znamensky (@russoz)
short_description: Wrapper for C(django-admin check)
version_added: 9.1.0
description:
  - This module is a wrapper for the execution of C(django-admin check).
extends_documentation_fragment:
  - community.general._attributes
  - community.general._django
options:
  databases:
    description:
      - Specify databases to run checks against.
      - If not specified, Django does not run database tests.
      - The parameter has been renamed to O(databases) in community.general 11.3.0. The old name is still available as an alias.
    type: list
    elements: str
    aliases: ["database"]
  deploy:
    description:
      - Include additional checks relevant in a deployment setting.
    type: bool
    default: false
  fail_level:
    description:
      - Message level that triggers failure.
      - Default is the Django default value. Check the documentation for the version being used.
    type: str
    choices: [CRITICAL, ERROR, WARNING, INFO, DEBUG]
  tags:
    description:
      - Restrict checks to specific tags.
    type: list
    elements: str
  apps:
    description:
      - Restrict checks to specific applications.
      - Default is to check all applications.
    type: list
    elements: str
notes:
  - The outcome of the module is found in the common return values RV(ignore:stdout), RV(ignore:stderr), RV(ignore:rc).
  - The module fails if RV(ignore:rc) is not zero.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
"""

EXAMPLES = r"""
- name: Check the entire project
  community.general.django_check:
    settings: myproject.settings

- name: Create the project using specific databases
  community.general.django_check:
    database:
      - somedb
      - myotherdb
    settings: fancysite.settings
    pythonpath: /home/joedoe/project/fancysite
    venv: /home/joedoe/project/fancysite/venv
"""

RETURN = r"""
run_info:
  description: Command-line execution information.
  type: dict
  returned: success and C(verbosity) >= 3
version:
  description: Version of Django.
  type: str
  returned: always
  sample: 5.1.2
  version_added: 10.0.0
"""

from ansible_collections.community.general.plugins.module_utils._django import DjangoModuleHelper


class DjangoCheck(DjangoModuleHelper):
    module = dict(
        argument_spec=dict(
            databases=dict(type="list", elements="str", aliases=["database"]),
            deploy=dict(type="bool", default=False),
            fail_level=dict(type="str", choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
            tags=dict(type="list", elements="str"),
            apps=dict(type="list", elements="str"),
        ),
        supports_check_mode=True,
    )
    django_admin_cmd = "check"
    django_admin_arg_order = "database_stacked_dash deploy fail_level tags apps"

    def __init_module__(self):
        self.vars.set("database_stacked_dash", self.vars.databases, output=False)


def main():
    DjangoCheck.execute()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The module file for cp_mgmt_hosts
"""

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from ansible.plugins.action import ActionBase
from ansible.module_utils.connection import Connection

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import (
    utils,
)
from ansible_collections.check_point.mgmt.plugins.module_utils.checkpoint import (
    CheckPointRequest,
    map_params_to_obj,
    map_obj_to_params,
    sync_show_params_with_add_params,
    remove_unwanted_key,
    contains_show_identifier_param,
)
from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
    AnsibleArgSpecValidator,
)
from ansible_collections.check_point.mgmt.plugins.modules.cp_mgmt_hosts import (
    DOCUMENTATION,
)


class ActionModule(ActionBase):
    """action module"""

    def __init__(self, *args, **kwargs):
        super(ActionModule, self).__init__(*args, **kwargs)
        self._result = None
        self.api_call_object = "host"
        self.api_call_object_plural_version = "hosts"
        self.module_return = "mgmt_hosts"
        self.key_transform = {
            "ip_address": "ip-address",
            "ipv4_address": "ipv4-address",
            "ipv6_address": "ipv6-address",
            "interfaces": "interfaces",
            "nat_settings": "nat-settings",
            "hide_behind": "hide-behind",
            "install_on": "install-on",
            "host_servers": "host-servers",
            "dns_server": "dns-server",
            "mail_server": "mail-server",
            "web_server": "web-server",
            "web_server_config": "web-server-config",
            "additional_ports": "additional-ports",
            "application_engines": "application-engines",
            "listen_standard_port": "listen-standard-port",
            "operating_system": "operating-system",
            "protected_by": "protected-by",
            "details_level": "details-level",
            "ignore_warnings": "ignore-warnings",
            "ignore_errors": "ignore-errors",
            "mask_length": "mask-length",
            "mask_length4": "mask-length4",
            "mask_length6": "mask-length4",
            "subnet_mask": "subnet-mask",
        }

    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=DOCUMENTATION,
            schema_format="doc",
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            self._result["failed"] = True
            self._result["msg"] = errors

    def search_for_existing_rules(
        self, conn_request, api_call_object, search_payload=None, state=None
    ):
        result = conn_request.post(api_call_object, state, data=search_payload)
        return result

    def search_for_resource_name(self, conn_request, payload):
        search_result = []
        round_trip = False

        search_payload = utils.remove_empties(payload)
        if search_payload.get("round_trip"):
            round_trip = True
        if search_payload.get("round_trip") is not None:
            del search_payload["round_trip"]
        search_payload = map_params_to_obj(search_payload, self.key_transform)
        if not contains_show_identifier_param(search_payload):
            search_result = self.search_for_existing_rules(
                conn_request,
                self.api_call_object_plural_version,
                search_payload,
                "gathered",
            )
            if search_result.get("code") == 200:
                search_result = search_result["response"]["objects"]
                return search_result
        else:
            search_result = self.search_for_existing_rules(
                conn_request, self.api_call_object, search_payload, "gathered"
            )
            if round_trip:
                search_result = sync_show_params_with_add_params(
                    search_result["response"], self.key_transform
                )
            elif search_result.get("code") and search_result["code"] == 200:
                search_result = search_result["response"]
            search_result = map_obj_to_params(
                search_result,
                self.key_transform,
                self.module_return,
            )
        if search_result.get("code") and search_result["code"] != 200:
            if (
                search_result.get("response")
                and "object_not_found" in search_result["response"]["code"]
                and "not found" in search_result["response"]["message"]
            ):
                search_result = {}
            elif "object_not_found" in search_result.get(
                "code"
            ) and "not found" in search_result.get("message"):
                search_result = {}
        return search_result

    def delete_module_api_config(self, conn_request, module_config_params):
        config = {}
        before = {}
        after = {}
        result = {}
        changed = False
        round_trip = False
        ckp_session_uid = None
        payload = utils.remove_empties(module_config_params)
        if payload.get("round_trip"):
            round_trip = True
            del payload["round_trip"]
        remove_from_response = ["uid", "read-only", "domain"]
        if round_trip:
            search_payload = {"name": payload["name"], "round_trip": True}
        else:
            search_payload = {"name": payload["name"]}
        search_result = self.search_for_resource_name(
            conn_request, search_payload
        )
        if search_result:
            if round_trip:
                search_result = remove_unwanted_key(
                    search_result, remove_from_response
                )
            before = search_result
        result = conn_request.post(
            self.api_call_object, self._task.args["state"], data=payload
        )
        if before:
            config.update({"before": before, "after": after})
        else:
            config.update({"before": before})
        if result.get("changed"):
            changed = True
            ckp_session_uid = result["checkpoint_session_uid"]
        return config, changed, ckp_session_uid

    def configure_module_api(self, conn_request, module_config_params):
        config = {}
        before = {}
        after = {}
        result = {}
        changed = False
        round_trip = False
        ckp_session_uid = None
        # Add to the THIS list for the value which needs to be excluded
        # from HAVE params when compared to WANT param like 'ID' can be
        # part of HAVE param but may not be part of your WANT param
        remove_from_response = ["uid", "read-only", "domain"]
        remove_from_set = []
        payload = utils.remove_empties(module_config_params)
        if payload.get("round_trip"):
            round_trip = True
            del payload["round_trip"]
        if payload.get("name"):
            if round_trip:
                search_payload = {"name": payload["name"], "round_trip": True}
            else:
                search_payload = {"name": payload["name"]}
            search_result = self.search_for_resource_name(
                conn_request, search_payload
            )
            if search_result:
                if round_trip:
                    search_result = remove_unwanted_key(
                        search_result, remove_from_response
                    )
                before = search_result
        payload = map_params_to_obj(payload, self.key_transform)
        delete_params = {
            "name": payload["name"],
        }
        result = conn_request.post(
            self.api_call_object,
            self._task.args["state"],
            data=payload,
            remove_keys=remove_from_set,
            delete_params=delete_params,
        )
        if result.get("changed"):
            if round_trip:
                search_result = sync_show_params_with_add_params(
                    result["response"], self.key_transform
                )
            else:
                search_result = map_obj_to_params(
                    result["response"],
                    self.key_transform,
                    self.module_return,
                )
            if round_trip:
                search_result = remove_unwanted_key(
                    search_result, remove_from_response
                )
            after = search_result
            ckp_session_uid = result["checkpoint_session_uid"]
            changed = True
        config.update({"before": before, "after": after})

        return config, changed, ckp_session_uid

    def run(self, tmp=None, task_vars=None):
        self._supports_check_mode = True
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._check_argspec()
        self._result["checkpoint_session_uid"] = None
        if self._result.get("failed"):
            return self._result
        conn = Connection(self._connection.socket_path)
        conn_request = CheckPointRequest(connection=conn, task_vars=task_vars)
        if self._task.args["state"] == "gathered":
            if self._task.args.get("config"):
                self._result["gathered"] = self.search_for_resource_name(
                    conn_request, self._task.args["config"]
                )
            else:
                self._result["gathered"] = self.search_for_resource_name(
                    conn_request, dict()
                )
        elif (
            self._task.args["state"] == "merged"
            or self._task.args["state"] == "replaced"
        ):
            if self._task.args.get("config"):
                (
                    self._result[self.module_return],
                    self._result["changed"],
                    self._result["checkpoint_session_uid"],
                ) = self.configure_module_api(
                    conn_request, self._task.args["config"]
                )
        elif self._task.args["state"] == "deleted":
            if self._task.args.get("config"):
                (
                    self._result[self.module_return],
                    self._result["changed"],
                    self._result["checkpoint_session_uid"],
                ) = self.delete_module_api_config(
                    conn_request, self._task.args["config"]
                )
        if self._result.get("checkpoint_session_uid") is None:
            del self._result["checkpoint_session_uid"]

        return self._result



#!/usr/bin/python
#
# Copyright (c) 2025 Klaas Demter (@Klaas-)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_recoveryservicesvaultconfig
version_added: "3.4.0"
short_description: Set Azure Backup Vault Config
description:
    - Get Azure Backup Vault Config
    - Most of the information here is also part of azure_rm_recoveryservicesvault
    - it's seperated out because of https://github.com/Azure/azure-rest-api-specs/issues/34218
    - A vault config has to exist, so you can not change the state, you can only change values, only if they are overwriteable
    - Supports diff mode
    - Check mode assumes all changes you make are legal and accepted by the API

options:
    vault_name:
        description:
            - The name of the Recovery Services Vault the vault config belongs to.
        required: true
        type: str
    resource_group:
        description:
            - The name of the resource group the vault is in.
        required: true
        type: str
    properties:
        description:
            - The recovery service vault config properties.
            - https://learn.microsoft.com/en-us/python/api/azure-mgmt-recoveryservicesbackup/
              azure.mgmt.recoveryservicesbackup.activestamp.models.backupresourcevaultconfig?view=azure-python
        type: dict
        suboptions:
            enhanced_security_state:
                description:
                    - Soft delete and security settings for hybrid workloads
                    - Enables soft delete, MFA and alert notifications for workloads running on premises.
                    - Refer to https://learn.microsoft.com/en-us/azure/backup/backup-azure-security-feature#minimum-version-requirements
                    - for minimum version requirements
                    - Can be reverted, unless set to AlwaysON
                type: str
                choices:
                    - Disabled
                    - Enabled
                    - AlwaysON
            soft_delete_feature_state:
                description:
                    - Enables soft delete for cloud workloads
                    - https://learn.microsoft.com/en-us/azure/backup/backup-azure-enhanced-soft-delete-about#whats-enhanced-soft-delete
                    - Can be reverted, unless set to AlwaysON
                type: str
                choices:
                    - Disabled
                    - Enabled
                    - AlwaysON
            soft_delete_retention_period_in_days:
                description:
                    - Duration of soft delete state
                    - minimum 14 days
                    - maximum 180 days
                type: int
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Klaas Demter (@Klaas-)
'''

EXAMPLES = '''
- name: Set backup vault config soft delete retention time
  azure.azcollection.azure_rm_recoveryservicesvaultconfig:
    vault_name: Vault_Name
    resource_group: Resource_Group_Name
    properties:
      soft_delete_retention_period_in_days: 17

- name: Set backup vault config soft delete to always on for cloud and hybrid workloads (not reversible except by deleting vault)
  azure.azcollection.azure_rm_recoveryservicesvaultconfig:
    vault_name: Vault_Name
    resource_group: Resource_Group_Name
    properties:
      soft_delete_feature_state: AlwaysON
      enhanced_security_state: AlwaysON
'''

RETURN = '''
id:
    description:
        - Id of specified backup vault config.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/providers/Microsoft.RecoveryServices/vaults/Vault_Name/backupconfig/vaultconfig"
name:
    description:
        - always returns "vaultconfig"
    type: str
    returned: always
    sample: "vaultconfig"
type:
    description:
        - Type of backup vault config.
    type: str
    returned: always
    sample: "Microsoft.RecoveryServices/vaults/backupconfig"
properties:
    description:
        - Attributes of the backup vault config.
    type: dict
    returned: always
    sample: {
            "enhanced_security_state": "Enabled",
            "is_soft_delete_feature_state_editable": true,
            "soft_delete_feature_state": "Enabled",
            "soft_delete_retention_period_in_days": 14
        }
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt

try:
    from azure.core.exceptions import ResourceNotFoundError

except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMBackupVaultConfig(AzureRMModuleBaseExt):
    """Information class for an Azure RM Backup Vault Configs"""

    def __init__(self):
        self.module_arg_spec = dict(
            vault_name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            properties=dict(
                type='dict',
                options=dict(
                    enhanced_security_state=dict(type='str', choices=['Disabled', 'Enabled', 'AlwaysON']),
                    soft_delete_feature_state=dict(type='str', choices=['Disabled', 'Enabled', 'AlwaysON']),
                    soft_delete_retention_period_in_days=dict(type='int')
                )
            )
        )

        self.vault_name = None
        self.resource_group = None
        self.properties = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            id=None,
            changed=False,
            diff=dict(
                before=None,
                after=None
            )
        )

        super(AzureRMBackupVaultConfig, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                       supports_check_mode=True,
                                                       supports_tags=False)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        existing_backup_vault_config_object = None
        existing_backup_vault_config = None

        existing_backup_vault_config_object = self.get_vault_config()

        if existing_backup_vault_config_object:
            existing_backup_vault_config = self.serialize_obj(existing_backup_vault_config_object, 'BackupResourceVaultConfigResource')

        if existing_backup_vault_config:
            self.log("Vault exists, update config")
            result_compare = dict(compare=[])

            # This is a workaround because PATCH does not support sending just the values you want to update
            # https://github.com/Azure/azure-rest-api-specs/issues/34530
            update_backup_vault_config = dict(
                properties=self.properties
            )
            merged_backup_vault_config = existing_backup_vault_config.copy()
            # This may need to be moved to a deep merge if future options are not flat
            merged_backup_vault_config['properties'] = {**existing_backup_vault_config['properties'], **update_backup_vault_config['properties']}

            if not self.default_compare({}, merged_backup_vault_config, existing_backup_vault_config, '', result_compare):
                self.log("Properties update existing vault config")
                if self.check_mode:
                    self.log("Check mode, not actually updating")
                    self.set_results(merged_backup_vault_config, True, existing_backup_vault_config)
                else:
                    self.log("Updating Vault config")
                    updated_backup_vault_config = self.serialize_obj(self.set_vault_config(merged_backup_vault_config), 'BackupResourceVaultConfigResource')
                    self.set_results(updated_backup_vault_config, True, existing_backup_vault_config)
            else:
                self.log("Properties do not update existing vault config")
                self.set_results(existing_backup_vault_config, False, existing_backup_vault_config)
        else:
            self.log("Vault does not exist")
            self.fail("Vault {0} does not exist in resource group {1}, failing".format(self.vault_name, self.resource_group))

        return self.results

    def get_vault_config(self):
        '''
        Gets the properties of the specified backup vault config.
        https://learn.microsoft.com/en-us/python/api/azure-mgmt-recoveryservicesbackup/azure.mgmt.recoveryservicesbackup.activestamp.aio.operations.backupresourcevaultconfigsoperations?view=azure-python

        :return: BackupResourceVaultConfigResource
        '''
        self.log("Checking for vault config of vault {0} in resource group {1}".format(self.vault_name, self.resource_group))

        vault_config = None

        try:
            vault_config = self.recovery_services_backup_client.backup_resource_vault_configs.get(vault_name=self.vault_name,
                                                                                                  resource_group_name=self.resource_group)
        except ResourceNotFoundError as ex:
            self.log("Could not find vault {0} in resource group {1}".format(self.vault_name, self.resource_group))

        return vault_config

    def set_vault_config(self, merged_backup_vault_config):
        '''
        Sets the properties of the specified backup vault config.
        https://learn.microsoft.com/en-us/python/api/azure-mgmt-recoveryservicesbackup/azure.mgmt.recoveryservicesbackup.activestamp.aio.operations.backupresourcevaultconfigsoperations?view=azure-python

        :return: BackupResourceVaultConfigResource
        '''
        self.log("Updating vault config of vault {0} in resource group {1}".format(self.vault_name, self.resource_group))

        vault_config = None

        try:
            vault_config = self.recovery_services_backup_client.backup_resource_vault_configs.update(vault_name=self.vault_name,
                                                                                                     resource_group_name=self.resource_group,
                                                                                                     parameters=merged_backup_vault_config)
        except ResourceNotFoundError as ex:
            self.log("Could not find vault {0} in resource group {1}".format(self.vault_name, self.resource_group))

        return vault_config

    def set_results(self, vault_config, changed=False, before=None):
        if vault_config:
            self.results['id'] = vault_config['id']
            # Not properly returned by sdk nor API
            # https://github.com/Azure/azure-rest-api-specs/issues/34530#issuecomment-2862950321
            # self.results['location'] = vault_config['location']
            self.results['name'] = vault_config['name']
            self.results['type'] = vault_config['type']
            self.results['properties'] = vault_config['properties']
            self.results['changed'] = changed
            self.results['diff']['before'] = before
            self.results['diff']['after'] = vault_config

        else:
            self.results['id'] = None
            self.results['name'] = None
            self.results['type'] = None


def main():
    """Main execution"""
    AzureRMBackupVaultConfig()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_wireless_ssids_firewall_isolation_allowlist_entries
short_description: Resource module for organizations _wireless _ssids _firewall _isolation
  _allowlist _entries
description:
  - Manage operations create, update and delete of the resource organizations _wireless
    _ssids _firewall _isolation _allowlist _entries.
  - Create isolation allow list MAC entry for this organization.
  - Destroy isolation allow list MAC entry for this organization.
  - Update isolation allow list MAC entry info.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  client:
    description: The client of allowlist.
    suboptions:
      mac:
        description: L2 Isolation mac address.
        type: str
    type: dict
  description:
    description: The description of mac address.
    type: str
  entryId:
    description: EntryId path parameter. Entry ID.
    type: str
  network:
    description: The Network that allowlist belongs to.
    suboptions:
      id:
        description: The ID of network.
        type: str
    type: dict
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  ssid:
    description: The SSID that allowlist belongs to.
    suboptions:
      number:
        description: The number of SSID.
        type: int
    type: dict
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for wireless createOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
    description: Complete reference of the createOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-wireless-ssids-firewall-isolation-allowlist-entry
  - name: Cisco Meraki documentation for wireless deleteOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
    description: Complete reference of the deleteOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!delete-organization-wireless-ssids-firewall-isolation-allowlist-entry
  - name: Cisco Meraki documentation for wireless updateOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
    description: Complete reference of the updateOrganizationWirelessSsidsFirewallIsolationAllowlistEntry
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!update-organization-wireless-ssids-firewall-isolation-allowlist-entry
notes:
  - SDK Method used are
    wireless.Wireless.create_organization_wireless_ssids_firewall_isolation_allowlist_entry,
    wireless.Wireless.delete_organization_wireless_ssids_firewall_isolation_allowlist_entry,
    wireless.Wireless.update_organization_wireless_ssids_firewall_isolation_allowlist_entry,
  - Paths used are
    post /organizations/{organizationId}/wireless/ssids/firewall/isolation/allowlist/entries,
    delete /organizations/{organizationId}/wireless/ssids/firewall/isolation/allowlist/entries/{entryId},
    put /organizations/{organizationId}/wireless/ssids/firewall/isolation/allowlist/entries/{entryId},
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_wireless_ssids_firewall_isolation_allowlist_entries:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    client:
      mac: A1:B2:C3:D4:E5:F6
    description: Example mac address
    network:
      id: N_123
    organizationId: string
    ssid:
      number: 2
- name: Delete by id
  cisco.meraki.organizations_wireless_ssids_firewall_isolation_allowlist_entries:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: absent
    entryId: string
    organizationId: string
- name: Update by id
  cisco.meraki.organizations_wireless_ssids_firewall_isolation_allowlist_entries:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    state: present
    client:
      mac: A1:B2:C3:D4:E5:F6
    description: Example mac address
    entryId: string
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "entryId": "string",
      "createdAt": "string",
      "lastUpdatedAt": "string",
      "description": "string",
      "client": {
        "mac": "string"
      },
      "ssid": {
        "id": "string",
        "name": "string",
        "number": 0
      },
      "network": {
        "id": "string",
        "name": "string"
      }
    }
"""



#!/usr/bin/python
#
# Copyright (c) 2025 Klaas Demter (@Klaas-)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_monitoractivitylogalerts
version_added: "3.7.0"
short_description: Create, update and delete Activity Log Alerts.
description:
    - Create, update and delete Activity Log Alerts.
    - >-
      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.aio.operations.activitylogalertsoperations?view=azure-python#azure-mgmt-monitor-v2020-10-01-aio-operations-activitylogalertsoperations-create-or-update)
    - U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.activitylogalertresource?view=azure-python)

options:
    name:
        description:
            - The name of the activity log alert you're creating, updating or deleting.
        required: true
        type: str
    resource_group:
        description:
            - The name of the resource group
        required: true
        type: str
    location:
        description:
            - Location of the activity log alert.
            - Should be optional, but is not optional on creation.
            - U(https://github.com/Azure/azure-rest-api-specs/issues/35342)
            - Activity log alerts are usually 'Global'. On creation module will default to 'Global' if unspecified.
        required: false
        type: str
    scopes:
        description:
            - A list of resource IDs that will be used as prefixes.
            - The alert will only apply to Activity Log events with resource IDs that fall under one of these prefixes.
            - This list must include at least one item for creation.
        type: list
        elements: str
    condition:
        description:
            - The condition that will cause this alert to activate.
            - >-
              U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.alertruleallofcondition?view=azure-python)
        type: dict
        suboptions:
            all_of:
                description:
                    - The list of Activity Log Alert rule conditions.
                    - >-
                      U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.alertruleanyoforleafcondition?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    field:
                        description:
                            - The name of the Activity Log event's field that this condition will examine.
                            - The possible values for this field are (case-insensitive) resourceId, category, caller, level,
                              operationName, resourceGroup, resourceProvider, status, subStatus, resourceType
                              or anything beginning with properties.
                        type: str
                    equals:
                        description:
                            - The value of the events field will be compared to this value (case-insensitive) to determine if the condition is met.
                        type: str
                    contains_any:
                        description:
                            - The value of the events field will be compared to the values in this
                              array (case-insensitive) to determine if the condition is met.
                        type: list
                        elements: str
                    any_of:
                        description:
                            - An Activity Log Alert rule condition that is met when at least one of its member leaf conditions are met.
                            - >-
                              U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.alertruleleafcondition?view=azure-python)
                        type: list
                        elements: dict
                        suboptions:
                            field:
                                description:
                                    - The name of the Activity Log events field that this condition will examine. The possible values for this field are
                                      (case-insensitive) resourceId, category, caller, level, operationName, resourceGroup, resourceProvider,
                                      status, subStatus, resourceType, or anything beginning with properties.
                                type: str
                            equals:
                                description:
                                    - The value of the event's field will be compared to this value (case-insensitive) to determine if the condition is met.
                                type: str
                            contains_any:
                                description:
                                    - The value of the event's field will be compared to the values in this array
                                      (case-insensitive) to determine if the condition is met.
                                type: list
                                elements: str
    actions:
        description:
            - The actions that will activate when the condition is met.
            - U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.actionlist?view=azure-python)
        type: dict
        suboptions:
            action_groups:
                description:
                    - The list of the Action Groups.
                    - U(https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.actiongroup?view=azure-python)
                type: list
                elements: dict
                suboptions:
                    action_group_id:
                        description:
                            - The resource ID of the Action Group.
                            - Must be set when setting an action group.
                        type: str
                    webhook_properties:
                        description:
                            - the dictionary of custom properties to include with the post operation. These data are appended to the webhook payload.
                        type: dict
    enabled:
        description:
            - Indicates whether this activity log alert is enabled.
            - If an activity log alert is not enabled, then none of its receivers will receive communications.
            - Server side default false -- should be true.
            - U(https://github.com/Azure/azure-rest-api-specs/issues/35343)
        type: bool
    description:
        description:
            - A description of this Activity Log Alert rule.
        type: str
    state:
        description:
            - State of the activity log alert.
            - Use C(present) for creating/updating a activity log alert.
            - Use C(absent) for deleting a activity log alert.
        default: present
        type: str
        choices:
            - present
            - absent
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Klaas Demter (@Klaas-)
'''

EXAMPLES = '''
- name: Add a activity log alert
  azure.azcollection.azure_rm_monitoractivitylogalerts:
    state: present
    resource_group: resource_group_name
    location: global
    name: activity_log_alert_resource_health_example
    scopes:
      - /subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name
    condition:
      all_of:
        - equals: ResourceHealth
          field: category
        - any_of:
            - equals: Unknown
              field: properties.currentHealthStatus
            - equals: Unavailable
              field: properties.currentHealthStatus
            - equals: Degraded
              field: properties.currentHealthStatus
        - any_of:
            - equals: Unknown
              field: properties.previousHealthStatus
            - equals: Unavailable
              field: properties.previousHealthStatus
            - equals: Degraded
              field: properties.previousHealthStatus
            - equals: Available
              field: properties.previousHealthStatus
        - any_of:
            - equals: PlatformInitiated
              field: properties.cause
    actions:
      action_groups:
        - action_group_id: "/subscriptions/<sub-id>/resourceGroups/resource_group_name/providers/microsoft.insights/actionGroups/action_group_name"

- name: Add tag to existing activity log alert
  azure.azcollection.azure_rm_monitoractivitylogalerts:
    state: present
    name: activity_log_alert_name
    resource_group: resource_group_name
    append_tags: true
    tags:
      ThisIsAnAddedExampleTag: ExampleValue

- name: Delete a activity log alert
  azure.azcollection.azure_rm_monitoractivitylogalerts:
    state: absent
    name: activity_log_alert_name
    resource_group: resource_group_name
'''

RETURN = '''
activitylogalert:
    description:
        - Details of the activity log alert
        - Is null on state==absent (activity log alert does not exist or will be deleted)
        - Assumes you make legal changes in check mode
    type: dict
    returned: always
    sample: {
        "actions": {
            "action_groups": [
                {
                    "action_group_id": \
"/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name/providers/microsoft.insights/actionGroups/action_group_name"
                }
            ]
        },
        "condition": {
            "all_of": [
                {
                    "equals": "ResourceHealth",
                    "field": "category"
                },
                {
                    "any_of": [
                        {
                            "equals": "Unknown",
                            "field": "properties.currentHealthStatus"
                        },
                        {
                            "equals": "Unavailable",
                            "field": "properties.currentHealthStatus"
                        },
                        {
                            "equals": "Degraded",
                            "field": "properties.currentHealthStatus"
                        }
                    ]
                },
                {
                    "any_of": [
                        {
                            "equals": "Unknown",
                            "field": "properties.previousHealthStatus"
                        },
                        {
                            "equals": "Unavailable",
                            "field": "properties.previousHealthStatus"
                        },
                        {
                            "equals": "Degraded",
                            "field": "properties.previousHealthStatus"
                        },
                        {
                            "equals": "Available",
                            "field": "properties.previousHealthStatus"
                        }
                    ]
                },
                {
                    "any_of": [
                        {
                            "equals": "PlatformInitiated",
                            "field": "properties.cause"
                        }
                    ]
                }
            ]
        },
        "enabled": false,
        "id": \
"/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name/providers/Microsoft.Insights/activityLogAlerts/activity_log_alert_name",
        "location": "global",
        "name": "activity_log_alert_name",
        "scopes": [
            "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/resource_group_name"
        ],
        "tags": {},
        "type": "Microsoft.Insights/ActivityLogAlerts"
    }
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt


try:
    from azure.core.exceptions import ResourceNotFoundError

except ImportError:
    # This is handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'ActivityLogAlert'

any_of_all_of_condition_spec = dict(
    field=dict(type='str'),
    equals=dict(type='str'),
    contains_any=dict(type='list', elements='str')
)

all_of_condition_spec = dict(
    field=dict(type='str'),
    equals=dict(type='str'),
    contains_any=dict(type='list', elements='str'),
    any_of=dict(type='list', elements='dict', options=any_of_all_of_condition_spec)
)
condition_spec = dict(
    all_of=dict(type='list', elements='dict', options=all_of_condition_spec)
)

action_groups_actions_spec = dict(
    action_group_id=dict(type='str'),
    webhook_properties=dict(type='dict')
)

actions_spec = dict(
    action_groups=dict(type='list', elements='dict', options=action_groups_actions_spec)
)


class AzureRMMonitorActivityLogAlert(AzureRMModuleBaseExt):
    """Information class for an Azure RM Activity Log Alerts"""

    def __init__(self):
        # https://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.activitylogalertresource?view=azure-python
        self.module_arg_spec = dict(
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            location=dict(type='str'),
            scopes=dict(type='list', elements='str'),
            condition=dict(type='dict', options=condition_spec),
            actions=dict(type='dict', options=actions_spec),
            enabled=dict(type='bool'),
            description=dict(type='str'),
            state=dict(type='str', choices=['present', 'absent'], default='present')
        )

        self.name = None
        self.resource_group = None
        self.location = None
        self.tags = None
        self.scopes = None
        self.condition = None
        self.actions = None
        self.enabled = None
        self.description = None
        self.state = None
        self.log_path = None
        self.log_mode = None

        self.results = dict(
            changed=False,
            activitylogalert=dict(),
            diff=dict(
                before=None,
                after=None
            )
        )

        super(AzureRMMonitorActivityLogAlert, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                             supports_check_mode=True,
                                                             supports_tags=True)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])

        # Defaults for variables
        result = None
        result_compare = dict(compare=[])
        before_dict = None

        # Get current activity log alert if it exists
        before_dict = self.get_activity_log_alert()

        # Create dict from input, without None values
        # hhttps://learn.microsoft.com/en-us/python/api/azure-mgmt-monitor/azure.mgmt.monitor.v2020_10_01.models.activitylogalertresource?view=azure-python
        activity_log_alert_template = {
            "location": self.location,
            "scopes": self.scopes,
            "condition": self.condition,
            "actions": self.actions,
            "enabled": self.enabled,
            "description": self.description
        }
        # Filter out all None values
        activity_log_alert_input = {key: value for key, value in activity_log_alert_template.items() if value is not None}

        # Create/Update if state==present
        if self.state == 'present':
            if before_dict is None:
                # activity log alert does not exist, create
                # On creation default to location of resource group unless otherwise noted in input variables
                if not self.location:
                    activity_log_alert_input['location'] = 'Global'
                # On creation input == what we send to api
                activity_log_alert_update = activity_log_alert_input
                # Needs to be extended by tags if set
                if self.tags:
                    activity_log_alert_update['tags'] = self.tags
                self.results['changed'] = True
                if self.check_mode:
                    # Check mode, skipping actual creation
                    pass
                else:
                    create_response = self.create_or_update(activity_log_alert_update)
            else:
                # activity log alert already exists, updating it
                # Dict for update is the union of existing object overwritten by input data
                activity_log_alert_update = before_dict | activity_log_alert_input

                # Enhanced with tags (special behaviour because of append_tags possibility)
                update_tags, update_tags_content = self.update_tags(before_dict.get('tags'))
                # Check if we need to update the activity log alert
                if update_tags or not self.default_compare({}, activity_log_alert_update, before_dict, '', result_compare):
                    activity_log_alert_update['tags'] = update_tags_content
                    # Need to create/update the activity log alert; changed -> True
                    self.results['changed'] = True
                    if self.check_mode:
                        # Check mode, skipping actual creation
                        pass
                    else:
                        create_response = self.create_or_update(activity_log_alert_update)

            if self.check_mode or not self.results['changed']:
                # When object was not updated or when running in check mode
                # assume activity_log_alert_update is resulting object
                result = activity_log_alert_update
            else:
                # otherwise take resulting new object from response of create call
                result = create_response

        # Delete activity log alert if state is absent and it exists
        # if it doesn't exist, it's already absent
        elif self.state == 'absent' and before_dict is not None:
            self.results['changed'] = True
            if self.check_mode:
                # do not delete in check mode
                pass
            else:
                self.delete()

        self.results['diff']['before'] = before_dict
        self.results['diff']['after'] = result
        self.results['activitylogalert'] = result

        return self.results

    def get_activity_log_alert(self):
        '''
        Gets the properties of the specified activity log alert.

        :return: List of Activity Log Alerts
        '''
        self.log("Checking if activity log alert {0} in resource group {1} is present".format(self.name,
                                                                                              self.resource_group))

        result = None
        response = None

        try:
            response = self.monitor_management_client_activity_log_alerts.activity_log_alerts.get(activity_log_alert_name=self.name,
                                                                                                  resource_group_name=self.resource_group)
        except ResourceNotFoundError as ex:
            self.log("Could not find activity log alert {0} in resource group {1}".format(self.name, self.resource_group))
        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def create_or_update(self, activity_log_alert_update):
        result = None
        response = None
        activity_log_alerts = self.monitor_management_client_activity_log_alerts.activity_log_alerts
        try:
            response = activity_log_alerts.create_or_update(resource_group_name=self.resource_group,
                                                            activity_log_alert_name=self.name,
                                                            activity_log_alert_rule=activity_log_alert_update,
                                                            logging_enable=False)
        except Exception as ex:
            self.fail("Error creating or update activity log alert {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        if response:
            result = self.serialize_obj(response, AZURE_OBJECT_CLASS)

        return result

    def delete(self):
        response = None
        try:
            response = self.monitor_management_client_activity_log_alerts.activity_log_alerts.delete(resource_group_name=self.resource_group,
                                                                                                     activity_log_alert_name=self.name)
        except Exception as ex:
            self.fail("Error deleting activity log alert {0} in resource group {1}: {2}".format(self.name, self.resource_group, str(ex)))

        return response


def main():
    """Main execution"""
    AzureRMMonitorActivityLogAlert()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (c) 2024 xuzhang3 (@xuzhang3), Fred-sun (@Fred-sun)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_capacityreservationgroup_info

version_added: "2.4.0"

short_description: Get or list the capacity reservation group

description:
    - Get or list the capacity reservation group.

options:
    resource_group:
        description:
            - Name of the resource group.
        type: str
    name:
        description:
            - Name of the capacity reservation group.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - xuzhang3 (@xuzhang3)
    - Fred-sun (@Fred-sun)

'''

EXAMPLES = '''
- name: Get facts of the capacity reservation group by name
  azure_rm_capacityreservationgroup_info:
    resource_group: myResourceGroup
    name: mycapacityreservationgroup

- name: List facts of the capacity reservation group by resource group
  azure_rm_capacityreservationgroup_info:
    resource_group: myResourceGroup

- name: List facts of the capacity reservation group by subscription and filter by tags
  azure_rm_capacityreservationgroup_info:
    tags:
      - testing
      - foo:bar
'''

RETURN = '''
capacity_reservation_group:
    description:
        - Current state of the Capacity Reservation Group.
    returned: always
    type: complex
    contains:
        id:
            description:
                - Resource ID.
            returned: always
            type: str
            sample: "/subscriptions/xxx-xxx/resourceGroups/testRG/providers/Microsoft.Compute/capacityReservationGroups/testname01"
        location:
            description:
                - The Geo-location where the resource lives.
            returned: always
            type: str
            sample: eastus
        name:
            description:
                - Resource name.
            returned: always
            type: str
            sample: testname01
        resource_group:
            description:
                - Name of resource group.
            type: str
            returned: always
            sample: myResourceGroup
        tags:
            description:
                - Resource tags.
            returned: always
            type: dict
            sample: { 'key': 'value' }
        zones:
            description:
                -  A list of all capacity reservation resource ids that belong to capacity reservation group.
            returned: always
            type: list
            sample: ['1', '2']
        type:
            description:
                - The resource type.
            type: str
            returned: always
            sample: "Microsoft.Compute/capacityReservationGroups/capacityReservations"
'''

try:
    from azure.core.exceptions import HttpResponseError
except Exception:
    # This is handled in azure_rm_common
    pass

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMCapacityReservationGroupInfo(AzureRMModuleBase):

    def __init__(self):

        self.module_arg_spec = dict(
            resource_group=dict(type='str'),
            name=dict(type='str'),
            tags=dict(type='list', elements='str')
        )

        self.results = dict(
            changed=False,
            capacity_reservation_groups=[]
        )

        self.resource_group = None
        self.name = None
        self.tags = None

        super(AzureRMCapacityReservationGroupInfo, self).__init__(self.module_arg_spec,
                                                                  supports_check_mode=True,
                                                                  supports_tags=False,
                                                                  facts_module=True)

    def exec_module(self, **kwargs):
        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        if self.name and self.resource_group:
            response = [self.get_by_name()]
        elif self.resource_group:
            response = self.list_by_resourcegroup()
        else:
            response = self.list_all()

        for item in response:
            if item is not None and self.has_tags(item.tags, self.tags):
                self.results['capacity_reservation_groups'].append(self.to_dict(item))

        return self.results

    def get_by_name(self):
        response = None
        try:
            response = self.compute_client.capacity_reservation_groups.get(self.resource_group, self.name)

        except HttpResponseError as exec:
            self.log("Failed to retrieves information about a capacity reservation group, Exception as {0}".format(exec))

        return response

    def list_by_resourcegroup(self):
        response = None
        try:
            response = self.compute_client.capacity_reservation_groups.list_by_resource_group(self.resource_group)
        except HttpResponseError as exec:
            self.log("Faild to list ssh public keys by resource group, exception as {0}".format(exec))
        return response

    def list_all(self):
        response = None
        try:
            response = self.compute_client.capacity_reservation_groups.list_by_subscription()
        except HttpResponseError as exc:
            self.fail("Failed to list all items - {0}".format(str(exc)))

        return response

    def to_dict(self, body):
        return dict(
            id=body.id,
            resource_group=self.resource_group,
            name=body.name,
            location=body.location,
            tags=body.tags,
            type=body.type,
            zones=body.zones,
        )


def main():
    AzureRMCapacityReservationGroupInfo()


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# Copyright (C) 2018 Huawei
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

###############################################################################
# Documentation
###############################################################################

DOCUMENTATION = r"""
module: hwc_network_vpc
description:
  - Represents an vpc resource.
short_description: Creates a Huawei Cloud VPC
author: Huawei Inc. (@huaweicloud)
requirements:
  - requests >= 2.18.4
  - keystoneauth1 >= 3.6.0
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  state:
    description:
      - Whether the given object should exist in VPC.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  timeouts:
    description:
      - The timeouts for each operations.
    type: dict
    default: {}
    suboptions:
      create:
        description:
          - The timeout for create operation.
        type: str
        default: '15m'
      update:
        description:
          - The timeout for update operation.
        type: str
        default: '15m'
      delete:
        description:
          - The timeout for delete operation.
        type: str
        default: '15m'
  name:
    description:
      - The name of vpc.
    type: str
    required: true
  cidr:
    description:
      - The range of available subnets in the VPC.
    type: str
    required: true
extends_documentation_fragment:
  - community.general._hwc
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Create a vpc
  community.general.hwc_network_vpc:
    identity_endpoint: "{{ identity_endpoint }}"
    user: "{{ user }}"
    password: "{{ password }}"
    domain: "{{ domain }}"
    project: "{{ project }}"
    region: "{{ region }}"
    name: "vpc_1"
    cidr: "192.168.100.0/24"
    state: present
"""

RETURN = r"""
id:
  description:
    - The ID of VPC.
  type: str
  returned: success
name:
  description:
    - The name of VPC.
  type: str
  returned: success
cidr:
  description:
    - The range of available subnets in the VPC.
  type: str
  returned: success
status:
  description:
    - The status of VPC.
  type: str
  returned: success
routes:
  description:
    - The route information.
  type: complex
  returned: success
  contains:
    destination:
      description:
        - The destination network segment of a route.
      type: str
      returned: success
    next_hop:
      description:
        - The next hop of a route. If the route type is peering, it provides VPC peering connection ID.
      type: str
      returned: success
enable_shared_snat:
  description:
    - Show whether the shared SNAT is enabled.
  type: bool
  returned: success
"""

###############################################################################
# Imports
###############################################################################

import re

from ansible_collections.community.general.plugins.module_utils._hwc_utils import (
    Config,
    HwcClientException,
    HwcClientException404,
    HwcModule,
    are_different_dicts,
    build_path,
    get_region,
    is_empty_value,
    navigate_value,
    wait_to_finish,
)

###############################################################################
# Main
###############################################################################


def main():
    """Main function"""

    module = HwcModule(
        argument_spec=dict(
            state=dict(default="present", choices=["present", "absent"], type="str"),
            timeouts=dict(
                type="dict",
                options=dict(
                    create=dict(default="15m", type="str"),
                    update=dict(default="15m", type="str"),
                    delete=dict(default="15m", type="str"),
                ),
                default=dict(),
            ),
            name=dict(required=True, type="str"),
            cidr=dict(required=True, type="str"),
        ),
        supports_check_mode=True,
    )
    config = Config(module, "vpc")

    state = module.params["state"]

    if (not module.params.get("id")) and module.params.get("name"):
        module.params["id"] = get_id_by_name(config)

    fetch = None
    link = self_link(module)
    # the link will include Nones if required format parameters are missed
    if not re.search("/None/|/None$", link):
        client = config.client(get_region(module), "vpc", "project")
        fetch = fetch_resource(module, client, link)
        if fetch:
            fetch = fetch.get("vpc")
    changed = False

    if fetch:
        if state == "present":
            expect = _get_editable_properties(module)
            current_state = response_to_hash(module, fetch)
            current = {"cidr": current_state["cidr"]}
            if are_different_dicts(expect, current):
                if not module.check_mode:
                    fetch = update(config, self_link(module))
                    fetch = response_to_hash(module, fetch.get("vpc"))
                changed = True
            else:
                fetch = current_state
        else:
            if not module.check_mode:
                delete(config, self_link(module))
                fetch = {}
            changed = True
    else:
        if state == "present":
            if not module.check_mode:
                fetch = create(config, "vpcs")
                fetch = response_to_hash(module, fetch.get("vpc"))
            changed = True
        else:
            fetch = {}

    fetch.update({"changed": changed})

    module.exit_json(**fetch)


def create(config, link):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    r = None
    try:
        r = client.post(link, resource_to_create(module))
    except HwcClientException as ex:
        msg = f"module(hwc_network_vpc): error creating resource, error: {ex}"
        module.fail_json(msg=msg)

    wait_done = wait_for_operation(config, "create", r)
    v = ""
    try:
        v = navigate_value(wait_done, ["vpc", "id"])
    except Exception as ex:
        module.fail_json(msg=str(ex))

    url = build_path(module, "vpcs/{op_id}", {"op_id": v})
    return fetch_resource(module, client, url)


def update(config, link):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    r = None
    try:
        r = client.put(link, resource_to_update(module))
    except HwcClientException as ex:
        msg = f"module(hwc_network_vpc): error updating resource, error: {ex}"
        module.fail_json(msg=msg)

    wait_for_operation(config, "update", r)

    return fetch_resource(module, client, link)


def delete(config, link):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    try:
        client.delete(link)
    except HwcClientException as ex:
        msg = f"module(hwc_network_vpc): error deleting resource, error: {ex}"
        module.fail_json(msg=msg)

    wait_for_delete(module, client, link)


def fetch_resource(module, client, link):
    try:
        return client.get(link)
    except HwcClientException as ex:
        msg = f"module(hwc_network_vpc): error fetching resource, error: {ex}"
        module.fail_json(msg=msg)


def get_id_by_name(config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")
    name = module.params.get("name")
    link = "vpcs"
    query_link = "?marker={marker}&limit=10"
    link += query_link
    not_format_keys = re.findall("={marker}", link)
    none_values = re.findall("=None", link)

    if not (not_format_keys or none_values):
        r = None
        try:
            r = client.get(link)
        except Exception:
            pass
        if r is None:
            return None
        r = r.get("vpcs", [])
        ids = [i.get("id") for i in r if i.get("name", "") == name]
        if not ids:
            return None
        elif len(ids) == 1:
            return ids[0]
        else:
            module.fail_json(msg="Multiple resources with same name are found.")
    elif none_values:
        module.fail_json(msg="Can not find id by name because url includes None.")
    else:
        p = {"marker": ""}
        ids = set()
        while True:
            r = None
            try:
                r = client.get(link.format(**p))
            except Exception:
                pass
            if r is None:
                break
            r = r.get("vpcs", [])
            if r == []:
                break
            for i in r:
                if i.get("name") == name:
                    ids.add(i.get("id"))
            if len(ids) >= 2:
                module.fail_json(msg="Multiple resources with same name are found.")

            p["marker"] = r[-1].get("id")

        return ids.pop() if ids else None


def self_link(module):
    return build_path(module, "vpcs/{id}")


def resource_to_create(module):
    params = dict()

    v = module.params.get("cidr")
    if not is_empty_value(v):
        params["cidr"] = v

    v = module.params.get("name")
    if not is_empty_value(v):
        params["name"] = v

    if not params:
        return params

    params = {"vpc": params}

    return params


def resource_to_update(module):
    params = dict()

    v = module.params.get("cidr")
    if not is_empty_value(v):
        params["cidr"] = v

    if not params:
        return params

    params = {"vpc": params}

    return params


def _get_editable_properties(module):
    return {
        "cidr": module.params.get("cidr"),
    }


def response_to_hash(module, response):
    """Remove unnecessary properties from the response.
    This is for doing comparisons with Ansible's current parameters.
    """
    return {
        "id": response.get("id"),
        "name": response.get("name"),
        "cidr": response.get("cidr"),
        "status": response.get("status"),
        "routes": VpcRoutesArray(response.get("routes", []), module).from_response(),
        "enable_shared_snat": response.get("enable_shared_snat"),
    }


def wait_for_operation(config, op_type, op_result):
    module = config.module
    op_id = ""
    try:
        op_id = navigate_value(op_result, ["vpc", "id"])
    except Exception as ex:
        module.fail_json(msg=str(ex))

    url = build_path(module, "vpcs/{op_id}", {"op_id": op_id})
    timeout = 60 * int(module.params["timeouts"][op_type].rstrip("m"))
    states = {
        "create": {
            "allowed": ["CREATING", "DONW", "OK"],
            "complete": ["OK"],
        },
        "update": {
            "allowed": ["PENDING_UPDATE", "DONW", "OK"],
            "complete": ["OK"],
        },
    }

    return wait_for_completion(url, timeout, states[op_type]["allowed"], states[op_type]["complete"], config)


def wait_for_completion(op_uri, timeout, allowed_states, complete_states, config):
    module = config.module
    client = config.client(get_region(module), "vpc", "project")

    def _refresh_status():
        r = None
        try:
            r = fetch_resource(module, client, op_uri)
        except Exception:
            return None, ""

        status = ""
        try:
            status = navigate_value(r, ["vpc", "status"])
        except Exception:
            return None, ""

        return r, status

    try:
        return wait_to_finish(complete_states, allowed_states, _refresh_status, timeout)
    except Exception as ex:
        module.fail_json(msg=str(ex))


def wait_for_delete(module, client, link):
    def _refresh_status():
        try:
            client.get(link)
        except HwcClientException404:
            return True, "Done"

        except Exception:
            return None, ""

        return True, "Pending"

    timeout = 60 * int(module.params["timeouts"]["delete"].rstrip("m"))
    try:
        return wait_to_finish(["Done"], ["Pending"], _refresh_status, timeout)
    except Exception as ex:
        module.fail_json(msg=str(ex))


class VpcRoutesArray:
    def __init__(self, request, module):
        self.module = module
        if request:
            self.request = request
        else:
            self.request = []

    def to_request(self):
        items = []
        for item in self.request:
            items.append(self._request_for_item(item))
        return items

    def from_response(self):
        items = []
        for item in self.request:
            items.append(self._response_from_item(item))
        return items

    def _request_for_item(self, item):
        return {"destination": item.get("destination"), "nexthop": item.get("next_hop")}

    def _response_from_item(self, item):
        return {"destination": item.get("destination"), "next_hop": item.get("nexthop")}


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipa_sudorule
author: Thomas Krahn (@Nosmoht)
short_description: Manage FreeIPA sudo rule
description:
  - Add, modify or delete sudo rule within IPA server using IPA API.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  cn:
    description:
      - Canonical name.
      - Can not be changed as it is the unique identifier.
    required: true
    aliases: ['name']
    type: str
  cmdcategory:
    description:
      - Command category the rule applies to.
    choices: ['all']
    type: str
  cmd:
    description:
      - List of commands assigned to the rule.
      - If an empty list is passed all commands are removed from the rule.
      - If option is omitted commands are not checked nor changed.
    type: list
    elements: str
  cmdgroup:
    description:
      - List of command groups assigned to the rule.
      - If an empty list is passed all command groups are removed from the rule.
      - If option is omitted command groups are not checked nor changed.
    type: list
    elements: str
    version_added: 2.0.0
  deny_cmd:
    description:
      - List of denied commands assigned to the rule.
      - If an empty list is passed all commands are removed from the rule.
      - If option is omitted commands are not checked nor changed.
    type: list
    elements: str
    version_added: 8.1.0
  deny_cmdgroup:
    description:
      - List of denied command groups assigned to the rule.
      - If an empty list is passed all command groups are removed from the rule.
      - If option is omitted command groups are not checked nor changed.
    type: list
    elements: str
    version_added: 8.1.0
  description:
    description:
      - Description of the sudo rule.
    type: str
  host:
    description:
      - List of hosts assigned to the rule.
      - If an empty list is passed all hosts are removed from the rule.
      - If option is omitted hosts are not checked nor changed.
      - Option O(hostcategory) must be omitted to assign hosts.
    type: list
    elements: str
  hostcategory:
    description:
      - Host category the rule applies to.
      - If V(all) is passed one must omit O(host) and O(hostgroup).
      - Option O(host) and O(hostgroup) must be omitted to assign V(all).
    choices: ['all']
    type: str
  hostgroup:
    description:
      - List of host groups assigned to the rule.
      - If an empty list is passed all host groups are removed from the rule.
      - If option is omitted host groups are not checked nor changed.
      - Option O(hostcategory) must be omitted to assign host groups.
    type: list
    elements: str
  runasextusers:
    description:
      - List of external RunAs users.
    type: list
    elements: str
    version_added: 2.3.0
  runasusercategory:
    description:
      - RunAs User category the rule applies to.
    choices: ['all']
    type: str
  runasgroupcategory:
    description:
      - RunAs Group category the rule applies to.
    choices: ['all']
    type: str
  sudoopt:
    description:
      - List of options to add to the sudo rule.
    type: list
    elements: str
  user:
    description:
      - List of users assigned to the rule.
      - If an empty list is passed all users are removed from the rule.
      - If option is omitted users are not checked nor changed.
    type: list
    elements: str
  usercategory:
    description:
      - User category the rule applies to.
    choices: ['all']
    type: str
  usergroup:
    description:
      - List of user groups assigned to the rule.
      - If an empty list is passed all user groups are removed from the rule.
      - If option is omitted user groups are not checked nor changed.
    type: list
    elements: str
  state:
    description: State to ensure.
    default: present
    choices: ['absent', 'disabled', 'enabled', 'present']
    type: str
extends_documentation_fragment:
  - community.general._ipa.documentation
  - community.general._ipa.connection_notes
  - community.general._attributes
"""

EXAMPLES = r"""
- name: Ensure sudo rule is present that's allows all every body to execute any command on any host without being asked
    for a password.
  community.general.ipa_sudorule:
    name: sudo_all_nopasswd
    cmdcategory: all
    description: Allow to run every command with sudo without password
    hostcategory: all
    sudoopt:
      - '!authenticate'
    usercategory: all
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure user group developers can run every command on host group db-server as well as on host db01.example.com.
  community.general.ipa_sudorule:
    name: sudo_dev_dbserver
    description: Allow developers to run every command with sudo on all database server
    cmdcategory: all
    host:
      - db01.example.com
    hostgroup:
      - db-server
    sudoopt:
      - '!authenticate'
    usergroup:
      - developers
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure user group operations can run any commands that is part of operations-cmdgroup on any host as user root.
  community.general.ipa_sudorule:
    name: sudo_operations_all
    description: Allow operators to run any commands that is part of operations-cmdgroup on any host as user root.
    cmdgroup:
      - operations-cmdgroup
    hostcategory: all
    runasextusers:
      - root
    sudoopt:
      - '!authenticate'
    usergroup:
      - operators
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
"""

RETURN = r"""
sudorule:
  description: Sudorule as returned by IPA.
  returned: always
  type: dict
"""

import traceback

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ipa import IPAClient, ipa_argument_spec
from ansible_collections.community.general.plugins.module_utils._version import LooseVersion


class SudoRuleIPAClient(IPAClient):
    def __init__(self, module, host, port, protocol):
        super().__init__(module, host, port, protocol)

    def sudorule_find(self, name):
        return self._post_json(method="sudorule_find", name=None, item={"all": True, "cn": name})

    def sudorule_add(self, name, item):
        return self._post_json(method="sudorule_add", name=name, item=item)

    def sudorule_add_runasuser(self, name, item):
        return self._post_json(method="sudorule_add_runasuser", name=name, item={"user": item})

    def sudorule_remove_runasuser(self, name, item):
        return self._post_json(method="sudorule_remove_runasuser", name=name, item={"user": item})

    def sudorule_mod(self, name, item):
        return self._post_json(method="sudorule_mod", name=name, item=item)

    def sudorule_del(self, name):
        return self._post_json(method="sudorule_del", name=name)

    def sudorule_add_option(self, name, item):
        return self._post_json(method="sudorule_add_option", name=name, item=item)

    def sudorule_add_option_ipasudoopt(self, name, item):
        return self.sudorule_add_option(name=name, item={"ipasudoopt": item})

    def sudorule_remove_option(self, name, item):
        return self._post_json(method="sudorule_remove_option", name=name, item=item)

    def sudorule_remove_option_ipasudoopt(self, name, item):
        return self.sudorule_remove_option(name=name, item={"ipasudoopt": item})

    def sudorule_add_host(self, name, item):
        return self._post_json(method="sudorule_add_host", name=name, item=item)

    def sudorule_add_host_host(self, name, item):
        return self.sudorule_add_host(name=name, item={"host": item})

    def sudorule_add_host_hostgroup(self, name, item):
        return self.sudorule_add_host(name=name, item={"hostgroup": item})

    def sudorule_remove_host(self, name, item):
        return self._post_json(method="sudorule_remove_host", name=name, item=item)

    def sudorule_remove_host_host(self, name, item):
        return self.sudorule_remove_host(name=name, item={"host": item})

    def sudorule_remove_host_hostgroup(self, name, item):
        return self.sudorule_remove_host(name=name, item={"hostgroup": item})

    def sudorule_add_allow_command(self, name, item):
        return self._post_json(method="sudorule_add_allow_command", name=name, item={"sudocmd": item})

    def sudorule_add_allow_command_group(self, name, item):
        return self._post_json(method="sudorule_add_allow_command", name=name, item={"sudocmdgroup": item})

    def sudorule_add_deny_command(self, name, item):
        return self._post_json(method="sudorule_add_deny_command", name=name, item={"sudocmd": item})

    def sudorule_add_deny_command_group(self, name, item):
        return self._post_json(method="sudorule_add_deny_command", name=name, item={"sudocmdgroup": item})

    def sudorule_remove_allow_command(self, name, item):
        return self._post_json(method="sudorule_remove_allow_command", name=name, item=item)

    def sudorule_add_user(self, name, item):
        return self._post_json(method="sudorule_add_user", name=name, item=item)

    def sudorule_add_user_user(self, name, item):
        return self.sudorule_add_user(name=name, item={"user": item})

    def sudorule_add_user_group(self, name, item):
        return self.sudorule_add_user(name=name, item={"group": item})

    def sudorule_remove_user(self, name, item):
        return self._post_json(method="sudorule_remove_user", name=name, item=item)

    def sudorule_remove_user_user(self, name, item):
        return self.sudorule_remove_user(name=name, item={"user": item})

    def sudorule_remove_user_group(self, name, item):
        return self.sudorule_remove_user(name=name, item={"group": item})


def get_sudorule_dict(
    cmdcategory=None,
    description=None,
    hostcategory=None,
    ipaenabledflag=None,
    usercategory=None,
    runasgroupcategory=None,
    runasusercategory=None,
):
    data = {}
    if cmdcategory is not None:
        data["cmdcategory"] = cmdcategory
    if description is not None:
        data["description"] = description
    if hostcategory is not None:
        data["hostcategory"] = hostcategory
    if ipaenabledflag is not None:
        data["ipaenabledflag"] = ipaenabledflag
    if usercategory is not None:
        data["usercategory"] = usercategory
    if runasusercategory is not None:
        data["ipasudorunasusercategory"] = runasusercategory
    if runasgroupcategory is not None:
        data["ipasudorunasgroupcategory"] = runasgroupcategory
    return data


def category_changed(module, client, category_name, ipa_sudorule):
    if ipa_sudorule.get(category_name, None) == ["all"]:
        if not module.check_mode:
            # cn is returned as list even with only a single value.
            client.sudorule_mod(name=ipa_sudorule.get("cn")[0], item={category_name: None})
        return True
    return False


def ensure(module, client):
    state = module.params["state"]
    name = module.params["cn"]
    cmd = module.params["cmd"]
    cmdgroup = module.params["cmdgroup"]
    cmdcategory = module.params["cmdcategory"]
    deny_cmd = module.params["deny_cmd"]
    deny_cmdgroup = module.params["deny_cmdgroup"]
    host = module.params["host"]
    hostcategory = module.params["hostcategory"]
    hostgroup = module.params["hostgroup"]
    runasusercategory = module.params["runasusercategory"]
    runasgroupcategory = module.params["runasgroupcategory"]
    runasextusers = module.params["runasextusers"]

    ipa_version = client.get_ipa_version()
    if state in ["present", "enabled"]:
        if LooseVersion(ipa_version) < LooseVersion("4.9.10"):
            ipaenabledflag = "TRUE"
        else:
            ipaenabledflag = True
    else:
        if LooseVersion(ipa_version) < LooseVersion("4.9.10"):
            ipaenabledflag = "FALSE"
        else:
            ipaenabledflag = False

    sudoopt = module.params["sudoopt"]
    user = module.params["user"]
    usercategory = module.params["usercategory"]
    usergroup = module.params["usergroup"]

    module_sudorule = get_sudorule_dict(
        cmdcategory=cmdcategory,
        description=module.params["description"],
        hostcategory=hostcategory,
        ipaenabledflag=ipaenabledflag,
        usercategory=usercategory,
        runasusercategory=runasusercategory,
        runasgroupcategory=runasgroupcategory,
    )
    ipa_sudorule = client.sudorule_find(name=name)

    changed = False
    if state in ["present", "disabled", "enabled"]:
        if not ipa_sudorule:
            changed = True
            if not module.check_mode:
                ipa_sudorule = client.sudorule_add(name=name, item=module_sudorule)
        else:
            diff = client.get_diff(ipa_sudorule, module_sudorule)
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    if "hostcategory" in diff:
                        if ipa_sudorule.get("memberhost_host", None) is not None:
                            client.sudorule_remove_host_host(name=name, item=ipa_sudorule.get("memberhost_host"))
                        if ipa_sudorule.get("memberhost_hostgroup", None) is not None:
                            client.sudorule_remove_host_hostgroup(
                                name=name, item=ipa_sudorule.get("memberhost_hostgroup")
                            )

                    client.sudorule_mod(name=name, item=module_sudorule)

        if cmd is not None:
            changed = category_changed(module, client, "cmdcategory", ipa_sudorule) or changed
            if not module.check_mode:
                client.sudorule_add_allow_command(name=name, item=cmd)

        if cmdgroup is not None:
            changed = category_changed(module, client, "cmdcategory", ipa_sudorule) or changed
            if not module.check_mode:
                client.sudorule_add_allow_command_group(name=name, item=cmdgroup)

        if deny_cmd is not None:
            changed = category_changed(module, client, "cmdcategory", ipa_sudorule) or changed
            if not module.check_mode:
                client.sudorule_add_deny_command(name=name, item=deny_cmd)

        if deny_cmdgroup is not None:
            changed = category_changed(module, client, "cmdcategory", ipa_sudorule) or changed
            if not module.check_mode:
                client.sudorule_add_deny_command_group(name=name, item=deny_cmdgroup)

        if runasusercategory is not None:
            changed = category_changed(module, client, "iparunasusercategory", ipa_sudorule) or changed

        if runasgroupcategory is not None:
            changed = category_changed(module, client, "iparunasgroupcategory", ipa_sudorule) or changed

        if host is not None:
            changed = category_changed(module, client, "hostcategory", ipa_sudorule) or changed
            changed = (
                client.modify_if_diff(
                    name,
                    ipa_sudorule.get("memberhost_host", []),
                    host,
                    client.sudorule_add_host_host,
                    client.sudorule_remove_host_host,
                )
                or changed
            )

        if hostgroup is not None:
            changed = category_changed(module, client, "hostcategory", ipa_sudorule) or changed
            changed = (
                client.modify_if_diff(
                    name,
                    ipa_sudorule.get("memberhost_hostgroup", []),
                    hostgroup,
                    client.sudorule_add_host_hostgroup,
                    client.sudorule_remove_host_hostgroup,
                )
                or changed
            )
        if sudoopt is not None:
            # client.modify_if_diff does not work as each option must be removed/added by its own
            ipa_list = ipa_sudorule.get("ipasudoopt", [])
            module_list = sudoopt
            diff = list(set(ipa_list) - set(module_list))
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    for item in diff:
                        client.sudorule_remove_option_ipasudoopt(name, item)
            diff = list(set(module_list) - set(ipa_list))
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    for item in diff:
                        client.sudorule_add_option_ipasudoopt(name, item)

        if runasextusers is not None:
            ipa_sudorule_run_as_user = ipa_sudorule.get("ipasudorunasextuser", [])
            diff = list(set(ipa_sudorule_run_as_user) - set(runasextusers))
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    for item in diff:
                        client.sudorule_remove_runasuser(name=name, item=item)
            diff = list(set(runasextusers) - set(ipa_sudorule_run_as_user))
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    for item in diff:
                        client.sudorule_add_runasuser(name=name, item=item)

        if user is not None:
            changed = category_changed(module, client, "usercategory", ipa_sudorule) or changed
            changed = (
                client.modify_if_diff(
                    name,
                    ipa_sudorule.get("memberuser_user", []),
                    user,
                    client.sudorule_add_user_user,
                    client.sudorule_remove_user_user,
                )
                or changed
            )
        if usergroup is not None:
            changed = category_changed(module, client, "usercategory", ipa_sudorule) or changed
            changed = (
                client.modify_if_diff(
                    name,
                    ipa_sudorule.get("memberuser_group", []),
                    usergroup,
                    client.sudorule_add_user_group,
                    client.sudorule_remove_user_group,
                )
                or changed
            )
    else:
        if ipa_sudorule:
            changed = True
            if not module.check_mode:
                client.sudorule_del(name)

    return changed, client.sudorule_find(name)


def main():
    argument_spec = ipa_argument_spec()
    argument_spec.update(
        cmd=dict(type="list", elements="str"),
        cmdgroup=dict(type="list", elements="str"),
        cmdcategory=dict(type="str", choices=["all"]),
        cn=dict(type="str", required=True, aliases=["name"]),
        deny_cmd=dict(type="list", elements="str"),
        deny_cmdgroup=dict(type="list", elements="str"),
        description=dict(type="str"),
        host=dict(type="list", elements="str"),
        hostcategory=dict(type="str", choices=["all"]),
        hostgroup=dict(type="list", elements="str"),
        runasusercategory=dict(type="str", choices=["all"]),
        runasgroupcategory=dict(type="str", choices=["all"]),
        sudoopt=dict(type="list", elements="str"),
        state=dict(type="str", default="present", choices=["present", "absent", "enabled", "disabled"]),
        user=dict(type="list", elements="str"),
        usercategory=dict(type="str", choices=["all"]),
        usergroup=dict(type="list", elements="str"),
        runasextusers=dict(type="list", elements="str"),
    )
    module = AnsibleModule(
        argument_spec=argument_spec,
        mutually_exclusive=[
            ["cmdcategory", "cmd"],
            ["cmdcategory", "deny_cmd"],
            ["cmdcategory", "cmdgroup"],
            ["cmdcategory", "deny_cmdgroup"],
            ["hostcategory", "host"],
            ["hostcategory", "hostgroup"],
            ["usercategory", "user"],
            ["usercategory", "usergroup"],
        ],
        supports_check_mode=True,
    )

    client = SudoRuleIPAClient(
        module=module,
        host=module.params["ipa_host"],
        port=module.params["ipa_port"],
        protocol=module.params["ipa_prot"],
    )
    try:
        client.login(username=module.params["ipa_user"], password=module.params["ipa_pass"])
        changed, sudorule = ensure(module, client)
        module.exit_json(changed=changed, sudorule=sudorule)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Red Hat, Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
---
module: win_route
short_description: Add or remove a static route
description:
    - Add or remove a static route.
version_added: 2.7.0
options:
  destination:
    description:
      - Destination IP address in CIDR format (ip address/prefix length).
    type: str
    required: yes
  gateway:
    description:
        - The gateway used by the static route.
        - If C(gateway) is not provided it will be set to C(0.0.0.0).
    type: str
    default: 0.0.0.0
  metric:
    description:
        - Metric used by the static route.
    type: int
    default: 1
  state:
    description:
      - If C(absent), it removes a network static route.
      - If C(present), it adds a network static route.
    type: str
    choices: [ absent, present ]
    default: present
notes:
  - Works only with Windows 2012 R2 and newer.
author:
- Amit Weinstock (@amitosw15)
'''

EXAMPLES = r'''
---
- name: Add a network static route
  ansible.windows.win_route:
    destination: 192.168.2.10/32
    gateway: 192.168.1.1
    metric: 1
    state: present

- name: Remove a network static route
  ansible.windows.win_route:
    destination: 192.168.2.10/32
    state: absent
'''
RETURN = r'''
msg:
    description: A message describing the task result.
    returned: always
    type: str
    sample: "Route added"
'''
